"""Tests for graphein.protein.features.nodes.geometry"""

# Graphein
# Author: Arian Jamasb <arian@jamasb.io>, Eric Ma
# License: MIT
# Project Website: https://github.com/a-r-j/graphein
# Code Repository: https://github.com/a-r-j/graphein

import operator
from functools import partial

import numpy as np
import pytest
from loguru import logger

from graphein.protein.config import ProteinGraphConfig
from graphein.protein.features.nodes.geometry import (
    add_beta_carbon_vector,
    add_sequence_neighbour_vector,
    add_sidechain_vector,
    add_virtual_beta_carbon_vector,
)
from graphein.protein.graphs import construct_graph


def test_add_beta_carbon_vector(caplog):
    config = ProteinGraphConfig(
        edge_construction_functions=[
            partial(add_beta_carbon_vector, scale=True)
        ],
    )
    g = construct_graph(pdb_code="1lds", config=config)

    raw_pdb = g.graph["raw_pdb_df"]
    for n, d in g.nodes(data=True):
        # Check that the node has the correct attributes
        assert "c_beta_vector" in d.keys()
        # Check the vector is of the correct dimensionality
        assert d["c_beta_vector"].shape == (3,)

        # check glycines are zero
        if d["residue_name"] == "GLY":
            np.testing.assert_equal(
                d["c_beta_vector"], np.array([0.0, 0.0, 0.0])
            )
        else:
            # Check scaled vector has norm close 1
            np.testing.assert_almost_equal(
                np.linalg.norm(d["c_beta_vector"]), 1.0
            )

    # Test unscaled vector
    config = ProteinGraphConfig(
        edge_construction_functions=[
            partial(add_beta_carbon_vector, scale=False)
        ],
    )
    g = construct_graph(pdb_code="1lds", config=config)

    for n, d in g.nodes(data=True):
        # check glycines are zero
        if d["residue_name"] == "GLY":
            np.testing.assert_equal(
                d["c_beta_vector"], np.array([0.0, 0.0, 0.0])
            )
        else:
            # Check the vector is pointing in the correct direction
            cb_true = np.array(
                raw_pdb.loc[
                    (raw_pdb.node_id == n) & (raw_pdb.atom_name == "CB")
                ][["x_coord", "y_coord", "z_coord"]]
            ).T.squeeze()
            np.testing.assert_almost_equal(
                cb_true, d["coords"] + d["c_beta_vector"]
            )
    # Test altloc handling
    g = construct_graph(config=config, pdb_code="6rew")
    for n, d in g.nodes(data=True):
        assert d["c_beta_vector"].shape == (3,)

    g = construct_graph(config=config, pdb_code="7w9w")
    for n, d in g.nodes(data=True):
        assert d["c_beta_vector"].shape == (3,)

    # Test handling of missing beta-carbons
    g = construct_graph(config=config, pdb_code="3se8")

    assert "H:CYS:104" in g.nodes
    for n, d in g.nodes(data=True):
        assert d["c_beta_vector"].shape == (3,)
        if n == "H:CYS:104":
            np.testing.assert_equal(
                d["c_beta_vector"], np.array([0.0, 0.0, 0.0])
            )


def test_add_sidechain_vector():
    config = ProteinGraphConfig(
        edge_construction_functions=[
            partial(add_sidechain_vector, scale=True)
        ],
    )
    g = construct_graph(pdb_code="1lds", config=config)

    for n, d in g.nodes(data=True):
        # Check that the node has the correct attributes
        assert "sidechain_vector" in d.keys()
        # Check the vector is of the correct dimensionality
        assert d["sidechain_vector"].shape == (3,)

        # check glycines are zero
        if d["residue_name"] == "GLY":
            np.testing.assert_equal(
                d["sidechain_vector"], np.array([0.0, 0.0, 0.0])
            )
        else:
            # Check scaled vector has norm close 1
            np.testing.assert_almost_equal(
                np.linalg.norm(d["sidechain_vector"]), 1.0
            )

    # Test unscaled vector
    config = ProteinGraphConfig(
        edge_construction_functions=[
            partial(add_sidechain_vector, scale=False)
        ],
    )
    g = construct_graph(pdb_code="1lds", config=config)

    for n, d in g.nodes(data=True):
        # check glycines are zero
        if d["residue_name"] == "GLY":
            np.testing.assert_equal(
                d["sidechain_vector"], np.array([0.0, 0.0, 0.0])
            )
        else:
            # Check the vector is pointing in the correct direction
            sc_true = np.array(
                g.graph["rgroup_df"]
                .groupby("node_id")
                .mean(numeric_only=True)
                .loc[n][["x_coord", "y_coord", "z_coord"]]
            )
            np.testing.assert_almost_equal(
                sc_true, d["coords"] + d["sidechain_vector"]
            )


def test_add_virtual_beta_carbon_vector():
    config = ProteinGraphConfig(
        edge_construction_functions=[
            partial(add_virtual_beta_carbon_vector, scale=True)
        ],
    )
    g = construct_graph(pdb_code="1lds", config=config)

    raw_pdb = g.graph["raw_pdb_df"]
    for n, d in g.nodes(data=True):
        # Check that the node has the correct attributes
        assert "virtual_c_beta_vector" in d.keys()
        # Check the vector is of the correct dimensionality
        assert d["virtual_c_beta_vector"].shape == (3,)

        # check glycines are not zero
        if d["residue_name"] == "GLY":
            np.testing.assert_array_compare(
                operator.__ne__,
                d["virtual_c_beta_vector"],
                np.array([0.0, 0.0, 0.0]),
            )
        else:
            # Check scaled vector has norm close 1
            np.testing.assert_almost_equal(
                np.linalg.norm(d["virtual_c_beta_vector"]), 1.0
            )

    # Test unscaled vector
    config = ProteinGraphConfig(
        edge_construction_functions=[
            partial(add_virtual_beta_carbon_vector, scale=False)
        ],
    )
    g = construct_graph(pdb_code="1lds", config=config)

    for n, d in g.nodes(data=True):
        # check glycines are not zero
        if d["residue_name"] == "GLY":
            np.testing.assert_array_compare(
                operator.__ne__,
                d["virtual_c_beta_vector"],
                np.array([0.0, 0.0, 0.0]),
            )

    # Test altloc handling
    g = construct_graph(config=config, pdb_code="6rew")
    for n, d in g.nodes(data=True):
        assert d["virtual_c_beta_vector"].shape == (3,)

    g = construct_graph(config=config, pdb_code="7w9w")
    for n, d in g.nodes(data=True):
        assert d["virtual_c_beta_vector"].shape == (3,)


@pytest.mark.parametrize("n_to_c", [True, False])
def test_add_sequence_neighbour_vector(n_to_c):
    config = ProteinGraphConfig(edge_construction_functions=[])
    g = construct_graph(pdb_code="1igt", config=config)
    add_sequence_neighbour_vector(g, n_to_c=n_to_c)

    key = "sequence_neighbour_vector_" + ("n_to_c" if n_to_c else "c_to_n")
    for n, d in g.nodes(data=True):
        # Check that the node has the correct attributes
        assert key in d.keys()
        # Check the vector is of the correct dimensionality
        assert d[key].shape == (3,)

        # check A insertions have non-zero backward vectors
        print(n, n_to_c, d[key])
        if n.endswith(":A") and not n_to_c:
            assert np.any(np.not_equal(d[key], [0.0, 0.0, 0.0]))



"""Functions for adding edges to Gene Regulatory Networks from various sources."""

# %%
# Graphein
# Author: Arian Jamasb <arian@jamasb.io>, Ramon Vinas
# License: MIT
# Project Website: https://github.com/a-r-j/graphein
# Code Repository: https://github.com/a-r-j/graphein
from typing import Callable, List, Optional

import networkx as nx
import pandas as pd
from loguru import logger as log

from graphein.grn.parse_regnetwork import RegNetwork_df
from graphein.grn.parse_trrust import TRRUST_df


def add_trrust_edges(
    G: nx.Graph, trrust_filtering_funcs: Optional[List[Callable]] = None
) -> nx.Graph:
    """
    Adds edges from TRRUST to GRNGraph

    :param G: Graph to edges to (populated with gene_id nodes)
    :type G: nx.Graph
    :param trrust_filtering_funcs: List of functions to apply to TRRUST
        dataframe as pre-processing prior to graph constructions. Defaults to
        ``None``.
    :type trrust_filtering_funcs: List[Callable], optional
    :return: nx.Graph GRNGraph with TRRUST regulatory interactions added
        as edges
    :rtype: nx.Graph
    """
    G.graph["sources"].append("trrust")
    G.graph["trrust_df"] = TRRUST_df(
        G.graph["gene_list"],
        filtering_funcs=trrust_filtering_funcs,
    )
    add_interacting_genes(G, df=G.graph["trrust_df"], kind="trrust")

    return G


def add_regnetwork_edges(
    G: nx.Graph, regnetwork_filtering_funcs: Optional[List[Callable]] = None
) -> nx.Graph:
    """
    Adds edges from RegNetwork to GRNGraph

    :param G: Graph to edges to (populated with gene_id nodes).
    :type G: nx.Graph
    :param kwargs:  Additional parameters to pass to RegNetwork.
    :return: nx.Graph GRNGraph with RegNetwork regulatory interactions added
        as edges.
    """
    G.graph["sources"].append("regnetwork")
    G.graph["regnetwork_df"] = RegNetwork_df(
        G.graph["gene_list"],
        filtering_funcs=regnetwork_filtering_funcs,
    )
    add_interacting_genes(G, df=G.graph["regnetwork_df"], kind="regnetwork")

    return G


def add_interacting_genes(
    G: nx.Graph, df: pd.DataFrame, kind: str
) -> nx.Graph:
    """
    Generic function for adding interaction edges to GRNGraph.

    :param G: GRNGraph to populate with edges.
    :type G: nx.Graph
    :param df: DataFrame containing edgelist.
    :type df: pd.DataFrame
    :param kind: name of interaction type.
    :type kind: str
    :returns: Graph with edges added.
    :rtype: nx.Graph
    """

    gene_1 = df["g1"].values
    gene_2 = df["g2"].values
    reg_type = df["regtype"].values

    interacting_genes = set(list(zip(gene_1, gene_2, reg_type)))

    for g1, g2, r in interacting_genes:
        if G.has_edge(g1, g2):
            G.edges[g1, g2]["kind"].add(kind)
            G.edges[g1, g2]["regtype"].add(r)
        else:
            G.add_edge(g1, g2, kind={kind}, regtype={r})
    log.debug(f"Added {len(df)} {kind} interaction edges")

    return G



"""Functions for featurising RNA Structure Graphs."""

# %%
# Graphein
# Author: Arian Jamasb <arian@jamasb.io>
# License: MIT
# Project Website: https://github.com/a-r-j/graphein
# Code Repository: https://github.com/a-r-j/graphein

from typing import Any, Dict

from graphein.rna.constants import RNA_ATOM_BOND_STATE, RNA_ATOMIC_RADII


def add_atomic_radii(n: str, d: Dict[str, Any]) -> float:
    """Add atomic radii to nodes based on values provided in:

    Structures of the Molecular Components in DNA and RNA with Bond
    Lengths Interpreted as Sums of Atomic Covalent Radii
    Raji Heyrovska

    Atoms in the RNA structure are mapped to their bond states
    (:const:`~graphein.rna.constants.RNA_ATOM_BOND_STATE`),
    which are in turn mapped to the corresponding atomic radii
    (:const:`~graphein.rna.constants.RNA_ATOMIC_RADII`).

    :param n: The node to add the atomic radius to.
        Unused, the argument is provided to retain a consistent function
        signature.
    :type n: str
    :param d: The node data.
    :type d: Dict[str, Any]
    :return: The atomic radius of the node.
    :rtype: float
    """

    base = d["residue_name"]
    atom = d["atom_type"]
    radius = RNA_ATOMIC_RADII[RNA_ATOM_BOND_STATE[base][atom]]
    d["atomic_radius"] = radius
    return radius



import torch
from torchmetrics import Metric

from ...protein.tensor.types import AtomTensor, CoordTensor


def gdt(
    y: Union[CoordTensor, AtomTensor],
    y_hat: Union[CoordTensor, AtomTensor],
    ca_only: bool = True,
    cutoff: float = 4,
    ts: bool = True,
) -> torch.Tensor:
    """Global Distance Deviation Test metric (GDDT).

    https://en.wikipedia.org/wiki/Global_distance_test


    The GDT score is calculated as the largest set of amino acid residues'
    alpha carbon atoms in the model structure falling within a defined
    distance cutoff of their position in the experimental structure, after
    iteratively superimposing the two structures. By the original design the
    GDT algorithm calculates 20 GDT scores, i.e. for each of 20 consecutive distance
    cutoffs (``0.5 Å, 1.0 Å, 1.5 Å, ... 10.0 Å``). For structure similarity assessment
    it is intended to use the GDT scores from several cutoff distances, and scores
    generally increase with increasing cutoff. A plateau in this increase may
    indicate an extreme divergence between the experimental and predicted structures,
    such that no additional atoms are included in any cutoff of a reasonable distance.
    The conventional GDT_TS total score in CASP is the average result of cutoffs at
    ``1``, ``2``, ``4``, and ``8`` Å.

    :param y: Tensor of groundtruth (reference) atom positions.
    :type y: Union[graphein.protein.tensor.CoordTensor, graphein.protein.tensor.AtomTensor]
    :param y_hat: Tensor of atom positions.
    :type y_hat: Union[graphein.protein.tensor.CoordTensor, graphein.protein.tensor.AtomTensor]
    :param ca_only: Whether or not to consider only Ca positions. Default is ``True``.
    :type ca_only: bool
    :param cutoff: Custom threshold to use.
    :type cutoff: float
    :param ts: Whether or not to use "Total Score" mode, where the scores over the thresholds
        ``1, 2, 4, 8`` are averaged (as per CASP).
    :type ts: bool
    :returns: GDT score (torch.FloatTensor)
    :rtype: torch.Tensor
    """
    if y.ndim == 3:
        y = y[:, 1, :] if ca_only else y.reshape(-1, 3)
    if y_hat.ndim == 3:
        y_hat = y_hat[:, 1, :] if ca_only else y_hat.reshape(-1, 3)
    # Get distance between points
    dist = torch.norm(y - y_hat, dim=1)

    if not ts:
        # Return fraction of distances below cutoff
        return (dist < cutoff).sum() / dist.numel()
    # Return mean fraction of distances below cutoff for each cutoff (1, 2, 4, 8)
    count_1 = (dist < 1).sum() / dist.numel()
    count_2 = (dist < 2).sum() / dist.numel()
    count_4 = (dist < 4).sum() / dist.numel()
    count_8 = (dist < 8).sum() / dist.numel()
    return torch.mean(torch.tensor([count_1, count_2, count_4, count_8]))


class GDT_TS(Metric):
    def __init__(self):
        """Torchmetrics implementation of GDT_TS."""
        super().__init__()
        self.add_state("total", default=torch.tensor(0), dist_reduce_fx="sum")
        self.add_state(
            "correct", default=torch.tensor(0.0), dist_reduce_fx="sum"
        )
        higher_is_better = True
        full_state_update = True

    @property
    def higher_is_better(self):
        return True

    def update(
        self, preds: torch.Tensor, target: torch.Tensor, batch: torch.Tensor
    ):
        """Update method for metric.

        :param pred: Tensor of predictions.
        :type pred: torch.Tensor
        :param target: Tensor of target structures.
        :type target: torch.Tensor
        :param batch. Batch tensor, indicating which indices belong to which example in the batch.
            Assumes a PyTorch Geometric batching scheme.
        type batch: torch.Tensor.
        """
        y = unbatch(target, batch)
        y_hat = unbatch(preds, batch)

        for i, j in zip(y, y_hat):
            self.correct += gdt(i, j)
            self.total += 1

    def compute(self) -> float:
        return self.correct / self.total



"""Generate diffusion matrices based on NetworkX graph objects.

Author: Arian Jamasb & Eric Ma

Each of the functions here accepts a NetworkX graph object
and returns a 2D xarray.
These arrays can then be stacked to generate a diffusion tensor.
"""

from typing import Dict

import networkx as nx
import numpy as np
import pandas as pd
import xarray as xr

from graphein.protein.edges.distance import compute_distmat
from graphein.utils.utils import format_adjacency, generate_feature_dataframe


def identity_matrix(G: nx.Graph) -> xr.DataArray:
    """Return the identity diffusion matrix.

    This is nothing more than the identity matrix
    with diagonals of ``1``.

    :param G: Graph to retrieve identity diffusion matrix from
    :type G: nx.Graph
    :returns: Identity diffusion matrix
    :rtype: xr.DataArray
    """
    return format_adjacency(G, np.eye(len(G)), "identity")


def adjacency_matrix_power(
    G: nx.Graph,
    amat_kwargs: Dict = {},
    with_identity: bool = True,
    power: float = 1,
) -> xr.DataArray:
    """
    Return the matrix power of the adjacency matrix.

    :param amat_kwargs: Keyword arguments to configure NetworkX's ``adjacency_matrix`` function. Defaults to ``{}``
    :type amat_kwargs: Dict[str, Any]
    :param with_identity: Whether or not to add in the identity matrix to the adjacency matrix. Effectively adding "self loops". Defaults to True.
    :type with_identity: bool
    :param power: Matrix power to raise the adjacency matrix, defaults to 1.
    :type power: float
    :returns: the matrix power of the adjacency matrix
    :rtype: xr.DataArray
    """
    adjacency_matrix_kwargs = {"weight": "weight", "nodelist": None}
    adjacency_matrix_kwargs.update(amat_kwargs)
    amat = nx.adjacency_matrix(G, **adjacency_matrix_kwargs)
    if with_identity:
        amat = amat + np.eye(len(G))
    amat = np.linalg.matrix_power(amat, power)
    return format_adjacency(
        G,
        amat,
        f"adjacency_matrix_power_{power}",
    )


def inverse_distance_matrix(G: nx.Graph, power: float) -> xr.DataArray:
    """
    Return the inverse distance matrix.

    Using the coordinates present on the graph object,
    calculate the inverse power distance matrix.

    The resultant array will be symmetric,
    with entries [i, j] corresponding to 1 / (distance**power),
    where distance == euclidean distance between the (x, y, z) coordinates
    of the two graphs.
    Diagonals (entries [i, i]) are set to 0.

    :param G: NetworkX Graph object. Assumes that each node has ``x_coord``, ``y_coord``, and ``z_coord`` as node attributes.
    :type G: nx.Graph
    :param power: The power for the distance calculation.
    :type power: float
    :returns: the inverse distance matrix.
    :type: xr.DataArray
    """

    def extract_coords(n, d) -> pd.Series:
        """
        Extracts xyz coordinates from a node.

        :param n: node id
        :param d: node data
        :return: pd.Series containing x,y,z coordinates
        :rtype: pd.Series
        """
        coord_names = ("x_coord", "y_coord", "z_coord")
        return pd.Series(
            {coord_name: d[coord_name] for coord_name in coord_names}, name=n
        )

    coords = generate_feature_dataframe(G, funcs=[extract_coords])
    distmat = compute_distmat(coords).values

    I = np.eye(len(G))
    distmat = (distmat + I) ** power
    distmat = 1 / distmat
    distmat -= I
    return format_adjacency(
        G, distmat, f"inverse_distance_matrix_power_{power}"
    )



"""
Module that defines a setup function and publishes the package to PyPI.
Use the command `python setup.py upload`.
"""

import codecs
import contextlib
import io
import os
import re
import sys
from pprint import pprint
from shutil import rmtree
from typing import List

from setuptools import Command, find_packages, setup

VERSION = None
HERE = os.path.abspath(os.path.dirname(__file__))
NAME = "graphein"

# Import the PYPI README and use it as the long-description.
# Note: this will only work if "README.md" is present in your MANIFEST.in file!
try:
    with io.open(os.path.join(HERE, "README.md"), encoding="utf-8") as f:
        long_description = "\n" + f.read()
except FileNotFoundError:
    long_description = DESCRIPTION


with io.open(
    os.path.join(os.path.dirname(__file__), "graphein/__init__.py"),
    encoding="utf-8",
) as f:
    for l in f:
        if not l.startswith("__version__"):
            continue
        VERSION = l.split("=")[1].strip(" \"'\n")
        break
PROJECT_ROOT = os.path.dirname(os.path.realpath(__file__))


def read(*parts):
    # intentionally *not* adding an encoding option to open
    return codecs.open(os.path.join(HERE, *parts), "r").read()


def read_requirements(*parts):
    """
    Return requirements from parts.
    Given a requirements.txt (or similar style file),
    returns a list of requirements.
    Assumes anything after a single '#' on a line is a comment, and ignores
    empty lines.
    :param parts: list of filenames which contain the installation "parts",
        i.e. submodule-specific installation requirements
    :returns: A compiled list of requirements.
    """
    requirements = []
    for line in read(*parts).splitlines():
        new_line = re.sub(  # noqa: PD005
            r"(\s*)?#.*$",  # the space immediately before the
            # hash mark, the hash mark, and
            # anything that follows it
            "",  # replace with a blank string
            line,
        )
        new_line = re.sub(  # noqa: PD005
            r"-r.*$",  # link to another requirement file
            "",  # replace with a blank string
            new_line,
        )
        new_line = re.sub(  # noqa: PD005
            r"-e \..*$",  # link to editable install
            "",  # replace with a blank string
            new_line,
        )
        # print(line, "-->", new_line)
        if new_line:  # i.e. we have a non-zero-length string
            requirements.append(new_line)
    return requirements


INSTALL_REQUIRES = read_requirements(".requirements/base.in")
EXTRA_REQUIRES = {
    "dev": read_requirements(".requirements/dev.in"),
    "extras": read_requirements(".requirements/extras.in"),
}
# Add all requires
all_requires = []
for v in EXTRA_REQUIRES.values():
    all_requires.extend(v)
EXTRA_REQUIRES["all"] = set(all_requires)

pprint(EXTRA_REQUIRES)


class UploadCommand(Command):
    """Support setup.py upload."""

    description = "Build and publish the package."
    user_options: List = []

    @staticmethod
    def status(s):
        """Print things in bold."""
        print("\033[1m{0}\033[0m".format(s))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        """Publish package to PyPI."""
        with contextlib.suppress(OSError):
            self.status("Removing previous builds…")
            rmtree(os.path.join(HERE, "dist"))
        self.status("Building Source and Wheel (universal) distribution…")
        os.system(
            "{0} setup.py sdist bdist_wheel --universal".format(sys.executable)
        )

        self.status("Uploading the package to PyPI via Twine…")
        os.system("twine upload dist/*")

        self.status("Pushing git tags…")
        os.system("git tag v{0}".format(VERSION))
        os.system("git push --tags")

        sys.exit()


setup(
    name="graphein",
    version="1.7.8",
    description="Protein & Interactomic Graph Construction for Machine Learning",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Arian Jamasb",
    author_email="arian@jamasb.io",
    url="https://github.com/a-r-j/graphein",
    project_urls={
        "homepage": "https://github.com/a-r-j/graphein",
        "changelog": "https://github.com/a-r-j/graphein/blob/master/CHANGELOG.md",
        "issue": "https://github.com/a-r-j/graphein/issues",
        "documentation": "https://graphein.ai/",
    },
    packages=find_packages(),
    package_data={
        "": ["LICENSE.txt", "README.md", "requirements.txt", "*.csv"]
    },
    include_package_data=True,
    # install_requires=install_reqs,
    install_requires=INSTALL_REQUIRES,
    extras_require=EXTRA_REQUIRES,
    python_requires=">=3.7",
    license="MIT",
    platforms="any",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Development Status :: 5 - Production/Stable",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX",
        "Operating System :: Unix",
        "Operating System :: MacOS",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Topic :: Scientific/Engineering",
    ],
    # $ setup.py publish support.
    cmdclass={
        "upload": UploadCommand,
    },
    entry_points={
        "console_scripts": [
            "graphein = graphein.cli:main",
        ],
    },
)



"""ML utility Functions for working with graphs."""

# Graphein
# Author: Arian Jamasb <arian@jamasb.io>
# License: MIT

# Project Website: https://github.com/a-r-j/graphein
# Code Repository: https://github.com/a-r-j/graphein
from typing import Any, List

import networkx as nx


def add_labels_to_graph(
    graphs: List[nx.Graph], labels: List[Any], name: str
) -> List[nx.Graph]:
    """Adds labels to a graph.

    :param graphs: A list of graphs to add labels to.
    :type graphs: List[nx.Graph]
    :param labels: A list of labels to add to the graphs.
    :type labels: List[Any]
    :param name: The name to add the label under.
    :type name: str
    :return: A list of graphs with labels added.
    :rtype: List[nx.Graph]
    """
    for i, g in enumerate(graphs):
        g.graph[name] = labels[i]
    return graphs



# basic
import functools
import pathlib

# test
import pytest

from graphein.protein.config import DSSPConfig, ProteinGraphConfig
from graphein.protein.edges import distance as D
from graphein.protein.features.nodes import rsa
from graphein.protein.graphs import construct_graph
from graphein.protein.subgraphs import extract_surface_subgraph
from graphein.protein.utils import ProteinGraphConfigurationError

# graphein
from graphein.utils.dependencies import is_tool

# ---------- input ----------
pdb_path = (
    pathlib.Path(__file__).parent.parent.parent
    / "test_data"
    / "input_pdb_cryst1.pdb"
)
dssp_exe = "mkdssp"
RSA_THRESHOLD = 0.2

DSSP_AVAILABLE = is_tool("mkdssp")


@pytest.fixture
def dssp_graph():
    # ---------- graph config ----------
    params_to_change = {
        "granularity": "centroids",  # "atom", "CA", "centroids"
        "insertions": True,
        "edge_construction_functions": [
            # graphein.protein.edges.distance.add_peptide_bonds,
            D.add_distance_to_edges,
            D.add_hydrogen_bond_interactions,
            D.add_ionic_interactions,
            D.add_backbone_carbonyl_carbonyl_interactions,
            D.add_salt_bridges,
            # distance
            functools.partial(
                D.add_distance_threshold,
                long_interaction_threshold=4,
                threshold=4.5,
            ),
        ],
        "dssp_config": DSSPConfig(executable=dssp_exe),
        "graph_metadata_functions": [rsa],
    }
    config = ProteinGraphConfig(**params_to_change)
    # ---------- construct graph ----------
    return construct_graph(config=config, path=pdb_path, verbose=False)


# ---------- test: dssp DataFrame ----------
@pytest.mark.skipif(not DSSP_AVAILABLE, reason="DSSP not available")
def test_assert_nonempty_dssp_df(dssp_graph):
    """if not provided dssp version to dssp.add_dssp_df, will output an empty DataFrame"""
    if dssp_graph.graph["dssp_df"].empty:
        pytest.fail("DSSP dataframe is empty")


# ---------- test: surface subgraph nodes with insertion code ----------
@pytest.mark.skipif(not DSSP_AVAILABLE, reason="DSSP not available")
def test_extract_surface_subgraph_insertion_node(dssp_graph):
    """if not added insertion codes, will raise ProteinGraphConfigurationError"""
    try:
        # without the modification, the following line will raise
        # ProteinGraphConfigurationError RSA not defined for all nodes (H:TYR:52:A).
        s_g = extract_surface_subgraph(dssp_graph, RSA_THRESHOLD)
    except ProteinGraphConfigurationError as e:
        pytest.fail(
            "extract_surface_subgraph raised ProteinGraphConfigurationError:\n{e}"
        )



"""Functions for making and parsing API calls to BIOGRID."""

# %%
# Graphein
# Author: Ramon Vinas, Arian Jamasb <arian@jamasb.io>
# License: MIT
# Project Website: https://github.com/a-r-j/graphein
# Code Repository: https://github.com/a-r-j/graphein

from typing import Dict, List, Union

import pandas as pd
import requests
from loguru import logger as log


def params_BIOGRID(
    params: Dict[str, Union[str, int, List[str], List[int]]], **kwargs
) -> Dict[str, Union[str, int]]:
    """
    Updates default parameters with user parameters for the method
    "interactions" of the BIOGRID API REST.

    See also https://wiki.thebiogrid.org/doku.php/biogridrest
    :param params: Dictionary of default parameters
    :type params: Dict[str, Union[str, int, List[str], List[int]]]
    :param kwargs: User parameters for the method "network" of the BIOGRID API
        REST. The key must start with "BIOGRID"
    :type kwargs: Dict[str, Union[str, int, List[str], List[int]]]
    :return: Dictionary of parameters
    :rtype: Dict[str, Union[str, int]]
    """
    fields = [
        "searchNames",  # If ‘true’, the interactor OFFICIAL_SYMBOL will be
        # examined for a match with the geneList.
        "max",  # Number of results to fetch
        "interSpeciesExcluded",  # If ‘true’, interactions with interactors from
        # different species will be excluded.
        "selfInteractionsExcluded",  # If ‘true’, interactions with one
        # interactor will be excluded.
        "evidenceList",  # Any interaction evidence with its Experimental System
        # in the list will be excluded from the results unless includeEvidence
        # is set to true.
        "includeEvidence",  # If set to true, any interaction evidence with its
        # Experimental System in the evidenceList will be included in the result
        "searchIds",  # If ‘true’, the interactor ENTREZ_GENE, ORDERED LOCUS and
        # SYSTEMATIC_NAME (orf) will be examined for a match with the geneList.
        "searchNames",  # If ‘true’, the interactor OFFICIAL_SYMBOL will be
        # examined for a match with the geneList.
        "searchSynonyms",  # If ‘true’, the interactor SYNONYMS will be examined
        # for a match with the geneList.
        "searchBiogridIds",  # If ‘true’, the entries in 'GENELIST' will be
        # compared to BIOGRID internal IDS which are provided in all Tab2
        # formatted files.
        "additionalIdentifierTypes",  # Identifier types on this list are
        # examined for a match with the geneList.
        "excludeGenes",  # If ‘true’, interactions containing genes in the
        # geneList will be excluded from the results.
        "includeInteractors",  # If ‘true’, in addition to interactions between
        # genes on the geneList, interactions will also be fetched which have
        # only one interactor on the geneList
        "includeInteractorInteractions",  # If ‘true’ interactions between the
        # geneList’s first order interactors will be included.
        "pubmedList",  # Interactions will be fetched whose Pubmed Id is/ is
        # not in this list, depending on the value of excludePubmeds.
        "excludePubmeds",  # If ‘false’, interactions with Pubmed ID in
        # pubmedList will be included in the results; if ‘true’ they will be
        # excluded.
        "htpThreshold",  # Interactions whose Pubmed ID has more than this
        # number of interactions will be excluded from the results. Ignored if
        # excludePubmeds is ‘false’.
        "throughputTag",  # If set to 'low or 'high', only interactions with
        # 'Low throughput' or 'High throughput' in the 'throughput' field
        # will be returned.
    ]
    for p in fields:
        kwarg_name = "BIOGRID_" + p
        if kwarg_name in kwargs:
            value = kwargs[kwarg_name]
            if type(value) is list:
                value = "|".join(value)
            params[p] = value
    return params


def parse_BIOGRID(
    protein_list: List[str],
    ncbi_taxon_id: Union[int, str, List[int], List[str]],
    paginate: bool = True,
    **kwargs,
) -> pd.DataFrame:
    """
    Makes BIOGRID API call and returns a source specific Pandas dataframe.

    See also [1] BIOGRID: https://wiki.thebiogrid.org/doku.php/biogridrest
    :param protein_list: Proteins to include in the graph
    :type protein_list: List[str]
    :param ncbi_taxon_id: NCBI taxonomy identifiers for the organism.
        Default is ``9606`` (Homo Sapiens).
    :type ncbi_taxon_id: Union[int, str, List[int], List[str]]
    :param paginate: boolean indicating whether to paginate the calls (for
        BIOGRID, the maximum number of rows per call is ``10000).`` Defaults to
        ``True``.
    :type paginate: bool
    :param kwargs: Parameters of the "interactions" method of the BIOGRID API
        REST, used to select the results. The parameter names are of the form
        ``"BIOGRID_<param>"``, where ``<param>`` is the name of the parameter.
        Information about these parameters can be found at [1].
    :type kwargs: Dict[str, Union[str, int, List[str], List[int]]]
    :return: Source specific Pandas dataframe.
    :rtype: pd.DataFrame
    """
    # Prepare call to BIOGRID API
    biogrid_api_url = "https://webservice.thebiogrid.org"
    method = "interactions"
    request_url = "/".join([biogrid_api_url, method])
    if type(ncbi_taxon_id) is list:
        ncbi_taxon_id = "|".join(str(t) for t in ncbi_taxon_id)
    params = {  # Default parameters
        "geneList": "|".join(protein_list),
        "accesskey": "c4ab86373e0bb921a878bb6d15ee4fb4",
        "taxId": ncbi_taxon_id,  # 9606 is human
        "format": "json",
        "max": 10000,  # Number of results to fetch
        "searchNames": "true",
        "includeInteractors": "false",  # Set to true to get any interaction
        # involving EITHER gene, set to false to get interactions between genes
        "selfInteractionsExcluded": "true",  # If ‘true’, interactions with one
        # interactor will be excluded
    }
    params = params_BIOGRID(params, **kwargs)

    # Call BIOGRID
    def make_call(
        request_url: str,
        params: Dict[str, Union[str, int]],
        start: int = 0,
        max: int = 10000,
        paginate: bool = paginate,
    ) -> pd.DataFrame:
        """
        Makes call to BIOGRID API.

        :param request_url: BIOGRID URL to make request
        :type request_url: str
        :param params: BIOGRID API parameters to use
        :type params: Dict[str, Union[str, int]]
        :param start: index in gene list to start from
        :type start: int
        :param max: number of genes to use in API call. Results are limited to
            10k per call
        :type max: int
        :param paginate: bool indicating whether or not to paginate calls.
            Above 10k calls this is required
        :type paginate: bool
        :return: pd.DataFrame containing BIOGRID_df API call output
        :rtype: pd.DataFrame
        """
        params["start"] = start
        response = requests.post(request_url, data=params)
        df = pd.read_json(response.text.strip()).transpose()

        # Maximum number of results is limited to 10k. Paginate to
        # retrieve everything
        if paginate and df.shape[0] == max:
            next_df = make_call(request_url, params, start + max, max)
            df = pd.concat([df, next_df])

        return df

    return make_call(
        request_url=request_url, params=params, start=0, max=params["max"]
    )


def filter_BIOGRID(df: pd.DataFrame, **kwargs) -> pd.DataFrame:
    """
    Filters results of the BIOGRID API call according to user kwargs.

    :param df: Source specific Pandas DataFrame (BIOGRID) with results of the
        API call.
    :type df: pd.DataFrame
    :param kwargs: User thresholds used to filter the results. The parameter
        names are of the form ``"BIOGRID_<param>"``, where ``<param>`` is the
        name of the parameter. All the parameters are numerical values.
    :type kwargs: Dict[str, Union[str, int, List[str], List[int]]]
    :return: Source specific Pandas DataFrame with filtered results.
    :rtype: pd.DataFrame
    """
    # Note: To filter BIOGRID interactions, use parameters from
    # https://wiki.thebiogrid.org/doku.php/biogridrest
    # TODO: Make sure that user can filter results of API call via the parameters.
    #       Otherwise implement filtering here.
    # TODO: Perhaps can filter by EXPERIMENTAL_SYSTEM (e.g. Co-fractionation)
    #       and EXPERIMENTAL_SYSTEM_TYPE (e.g. physical)
    return df


def standardise_BIOGRID(df: pd.DataFrame) -> pd.DataFrame:
    """
    Standardises BIOGRID DataFrame, e.g. puts everything into a common format.

    :param df: Source specific Pandas DataFrame.
    :type df: pd.DataFrame
    :return: Standardised DataFrame.
    :rtype: pd.DataFrame
    """
    if df.empty:
        return pd.DataFrame({"p1": [], "p2": [], "source": []})

    # Rename & delete columns
    df = df.rename(
        columns={"OFFICIAL_SYMBOL_A": "p1", "OFFICIAL_SYMBOL_B": "p2"}
    )
    df = df[["p1", "p2"]]

    # Add source column
    df["source"] = "BIOGRID"

    return df


def BIOGRID_df(
    protein_list: List[str],
    ncbi_taxon_id: Union[int, str, List[int], List[str]],
    **kwargs,
) -> pd.DataFrame:
    """
    Generates standardised DataFrame with BIOGRID protein-protein interactions,
    filtered according to user's input.

    :param protein_list: List of proteins (official symbol) that will be
        included in the PPI graph.
    :type protein_list: List[str]
    :param ncbi_taxon_id: NCBI taxonomy identifiers for the organism.
        ``9606`` corresponds to Homo Sapiens.
    :type ncbi_taxon_id: int
    :param kwargs:  Additional parameters to pass to BIOGRID API calls.
    :type kwargs: Union[int, str, List[int], List[str]]
    :return: Standardised DataFrame with BIOGRID interactions.
    :rtype: pd.DataFrame
    """
    df = parse_BIOGRID(
        protein_list=protein_list, ncbi_taxon_id=ncbi_taxon_id, **kwargs
    )
    df = filter_BIOGRID(df, **kwargs)
    df = standardise_BIOGRID(df)
    return df



"""Functions for working with Protein Structure Graphs."""

# %%
# Graphein
# Author: Arian Jamasb <arian@jamasb.io>, Eric Ma, Charlie Harris
# License: MIT
# Project Website: https://github.com/a-r-j/graphein
# Code Repository: https://github.com/a-r-j/graphein
from __future__ import annotations

import os
import traceback
from contextlib import nullcontext
from functools import partial
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import cpdb
import networkx as nx
import numpy as np
import pandas as pd
from biopandas.mmcif import PandasMmcif
from biopandas.mmtf import PandasMmtf
from biopandas.pdb import PandasPdb
from loguru import logger as log
from rich.progress import Progress
from tqdm.contrib.concurrent import process_map

from graphein.protein.config import GetContactsConfig, ProteinGraphConfig
from graphein.protein.edges.distance import (
    add_distance_to_edges,
    compute_distmat,
)
from graphein.protein.resi_atoms import BACKBONE_ATOMS, RESI_THREE_TO_1
from graphein.protein.subgraphs import extract_subgraph_from_chains
from graphein.protein.utils import (
    ProteinGraphConfigurationError,
    compute_rgroup_dataframe,
    filter_dataframe,
    get_protein_name_from_filename,
    three_to_one_with_mods,
)
from graphein.rna.constants import RNA_ATOMS
from graphein.utils.utils import (
    annotate_edge_metadata,
    annotate_graph_metadata,
    annotate_node_metadata,
    compute_edges,
)

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal


def subset_structure_to_rna(
    df: pd.DataFrame,
) -> pd.DataFrame:
    """
    Return a subset of atomic DataFrame that contains only certain atom names
    relevant for RNA structures.

    :param df: Protein Structure DataFrame to subset.
    :type df: pd.DataFrame
    :returns: Subsetted protein structure DataFrame.
    :rtype: pd.DataFrame
    """
    return filter_dataframe(
        df, by_column="atom_name", list_of_values=RNA_ATOMS, boolean=True
    )


def read_pdb_to_dataframe(
    path: Optional[Union[str, os.PathLike]] = None,
    pdb_code: Optional[str] = None,
    uniprot_id: Optional[str] = None,
    model_index: int = 1,
) -> pd.DataFrame:
    """
    Reads PDB file to ``PandasPDB`` object.

    Returns ``atomic_df``, which is a DataFrame enumerating all atoms and
    their cartesian coordinates in 3D space. Also contains associated metadata
    from the PDB file.

    :param path: path to PDB or MMTF file. Defaults to ``None``.
    :type path: str, optional
    :param pdb_code: 4-character PDB accession. Defaults to ``None``.
    :type pdb_code: str, optional
    :param uniprot_id: UniProt ID to build graph from AlphaFoldDB. Defaults to
        ``None``.
    :type uniprot_id: str, optional
    :param model_index: Index of model to read. Only relevant for structures
        containing ensembles. Defaults to ``1``.
    :type model_index: int, optional
    :returns: ``pd.DataFrame`` containing protein structure
    :rtype: pd.DataFrame
    """
    if pdb_code is None and path is None and uniprot_id is None:
        raise NameError(
            "One of pdb_code, path or uniprot_id must be specified!"
        )

    if path is not None:
        if isinstance(path, Path):
            path = os.fsdecode(path)
        if (
            path.endswith(".pdb")
            or path.endswith(".pdb.gz")
            or path.endswith(".ent")
        ):
            atomic_df = cpdb.parse(path)
        elif path.endswith(".mmtf") or path.endswith(".mmtf.gz"):
            atomic_df = PandasMmtf().read_mmtf(path)
            atomic_df = atomic_df.get_model(model_index)
            atomic_df = pd.concat(
                [atomic_df.df["ATOM"], atomic_df.df["HETATM"]]
            )
        elif (
            path.endswith(".cif")
            or path.endswith(".cif.gz")
            or path.endswith(".mmcif")
            or path.endswith(".mmcif.gz")
        ):
            atomic_df = PandasMmcif().read_mmcif(path)
            atomic_df = atomic_df.get_model(model_index)
            atomic_df = atomic_df.convert_to_pandas_pdb()
            atomic_df = pd.concat(
                [atomic_df.df["ATOM"], atomic_df.df["HETATM"]]
            )
        else:
            raise ValueError(
                f"File {path} must be either .pdb(.gz), .mmtf(.gz), .(mm)cif(.gz) or .ent, not {path.split('.')[-1]}"
            )
    elif uniprot_id is not None:
        atomic_df = cpdb.parse(uniprot_id=uniprot_id)
    else:
        atomic_df = cpdb.parse(pdb_code=pdb_code)

    if "model_idx" in atomic_df.columns:
        atomic_df = atomic_df.loc[atomic_df["model_idx"] == model_index]

    if len(atomic_df) == 0:
        raise ValueError(f"No model found for index: {model_index}")

    return atomic_df


def label_node_id(
    df: pd.DataFrame, granularity: str, insertions: bool = False
) -> pd.DataFrame:
    """Assigns a ``node_id`` column to the atomic dataframe. Node IDs are of the
    form: ``"<CHAIN>:<RESIDUE_NAME>:<RESIDUE_NUMBER>:<ATOM_NAME>"`` for atomic
    graphs or ``"<CHAIN>:<RESIDUE_NAME>:<RESIDUE_NUMBER>"`` for residue graphs.

    If ``insertions=True``, the insertion code will be appended to the end of
    the node_id (e.g. ``"<CHAIN>:<RESIDUE_NAME>:<RESIDUE_NUMBER>:<ATOM_NAME>:"``)

    :param df: Protein structure DataFrame.
    :type df: pd.DataFrame
    :param granularity: Granularity of graph. Atom-level,
        residue (e.g. ``CA``) or ``centroids``. See:
        :const:`~graphein.protein.config.GRAPH_ATOMS` and
        :const:`~graphein.protein.config.GRANULARITY_OPTS`.
    :type granularity: str
    :param insertions: Whether or not to include insertion codes in the node id.
        Default is ``False``.
    :type insertions: bool
    :return: Protein structure DataFrame with ``node_id`` column.
    :rtype: pd.DataFrame
    """
    df["node_id"] = (
        df["chain_id"].apply(str)
        + ":"
        + df["residue_name"]
        + ":"
        + df["residue_number"].apply(str)
    )

    if insertions:
        df["node_id"] = df["node_id"] + ":" + df["insertion"].apply(str)
        # Replace trailing : for non insertions
        df["node_id"] = df["node_id"].str.replace(":$", "", regex=True)
    # Add Alt Loc identifiers
    df["node_id"] = df["node_id"] + ":" + df["alt_loc"].apply(str)
    df["node_id"] = df["node_id"].str.replace(":$", "", regex=True)
    df["residue_id"] = df["node_id"]
    if granularity == "atom":
        df["node_id"] = df["node_id"] + ":" + df["atom_name"]
    elif granularity in {"rna_atom", "rna_centroid"}:
        df["node_id"] = (
            df["node_id"]
            + ":"
            + df["atom_number"].apply(str)
            + ":"
            + df["atom_name"]
        )
    return df


def deprotonate_structure(df: pd.DataFrame) -> pd.DataFrame:
    """Remove protons from PDB DataFrame.

    :param df: Atomic dataframe.
    :type df: pd.DataFrame
    :returns: Atomic dataframe with all ``element_symbol == "H" or "D" or "T"`` removed.
    :rtype: pd.DataFrame
    """
    log.debug(
        "Deprotonating protein. This removes H atoms from the pdb_df dataframe"
    )
    return filter_dataframe(
        df,
        by_column="element_symbol",
        list_of_values=["H", "D", "T"],
        boolean=False,
    )


def convert_structure_to_centroids(df: pd.DataFrame) -> pd.DataFrame:
    """Overwrite existing ``(x, y, z)`` coordinates with centroids of the amino
    acids.

    :param df: Pandas DataFrame protein structure to convert into a dataframe of
        centroid positions.
    :type df: pd.DataFrame
    :return: pd.DataFrame with atoms/residues positions converted into centroid
        positions.
    :rtype: pd.DataFrame
    """
    log.debug(
        "Converting dataframe to centroids. This averages XYZ coords of the \
            atoms in a residue"
    )

    centroids = calculate_centroid_positions(df)
    df = df.loc[df["atom_name"] == "CA"].reset_index(drop=True)
    df["x_coord"] = centroids["x_coord"]
    df["y_coord"] = centroids["y_coord"]
    df["z_coord"] = centroids["z_coord"]

    return df


def subset_structure_to_atom_type(
    df: pd.DataFrame, granularity: str
) -> pd.DataFrame:
    """
    Return a subset of atomic dataframe that contains only certain atom names.

    :param df: Protein Structure dataframe to subset.
    :type df: pd.DataFrame
    :returns: Subset protein structure dataframe.
    :rtype: pd.DataFrame
    """
    return filter_dataframe(
        df, by_column="atom_name", list_of_values=[granularity], boolean=True
    )


def remove_alt_locs(
    df: pd.DataFrame, keep: str = "max_occupancy"
) -> pd.DataFrame:
    """
    This function removes alternatively located atoms from PDB DataFrames
    (see https://proteopedia.org/wiki/index.php/Alternate_locations). Among the
    alternative locations the ones with the highest occupancies are left.

    :param df: Protein Structure dataframe to remove alternative located atoms
        from.
    :type df: pd.DataFrame
    :param keep: Controls how to remove altlocs. Default is ``"max_occupancy"``.
    :type keep: Literal["max_occupancy", "min_occupancy", "first", "last"]
    :return: Protein structure dataframe with alternative located atoms removed
    :rtype: pd.DataFrame
    """
    # Sort accordingly
    if keep == "max_occupancy":
        df = df.sort_values("occupancy")
        keep = "last"
    elif keep == "min_occupancy":
        df = df.sort_values("occupancy")
        keep = "first"
    elif keep == "exclude":
        keep = False

    # Filter
    duplicates = df.duplicated(
        subset=["chain_id", "residue_number", "atom_name", "insertion"],
        keep=keep,
    )
    df = df[~duplicates]

    # Unsort
    if keep in ["max_occupancy", "min_occupancy"]:
        df = df.sort_index()
    df = df.reset_index(drop=True)
    return df


def remove_insertions(
    df: pd.DataFrame, keep: Literal["first", "last"] = "first"
) -> pd.DataFrame:
    """
    This function removes insertions from PDB DataFrames.

    :param df: Protein Structure dataframe to remove insertions from.
    :type df: pd.DataFrame
    :param keep: Specifies which insertion to keep. Options are ``"first"`` or
        ``"last"``. Default is ``"first"``.
    :type keep: Literal["first", "last"]
    :return: Protein structure dataframe with insertions removed
    :rtype: pd.DataFrame
    """
    # Catches unnamed insertions
    duplicates = df.duplicated(
        subset=["chain_id", "residue_number", "atom_name", "alt_loc"],
        keep=keep,
    )
    df = df[~duplicates]

    return filter_dataframe(
        df, by_column="insertion", list_of_values=[""], boolean=True
    )


def filter_hetatms(
    df: pd.DataFrame, keep_hets: List[str]
) -> List[pd.DataFrame]:
    """Return hetatms of interest.

    :param df: Protein Structure dataframe to filter hetatoms from.
    :type df: pd.DataFrame
    :param keep_hets: List of hetero atom names to keep.
    :type keep_hets: List[str]
    :returns: Protein structure dataframe with heteroatoms removed
    :rtype: pd.DataFrame
    """
    return [df.loc[df["residue_name"] == hetatm] for hetatm in keep_hets]


def process_dataframe(
    protein_df: pd.DataFrame,
    atom_df_processing_funcs: Optional[List[Callable]] = None,
    hetatom_df_processing_funcs: Optional[List[Callable]] = None,
    granularity: str = "centroids",
    chain_selection: str = "all",
    insertions: bool = False,
    alt_locs: bool = False,
    deprotonate: bool = True,
    keep_hets: List[str] = [],
) -> pd.DataFrame:
    """
    Process ATOM and HETATM dataframes to produce singular dataframe used for
    graph construction.

    :param protein_df: Dataframe to process.
        Should be the object returned from
        :func:`~graphein.protein.graphs.read_pdb_to_dataframe`.
    :type protein_df: pd.DataFrame
    :param atom_df_processing_funcs: List of functions to process DataFrame.
        These must take in a DataFrame and return a DataFrame. Defaults to
        ``None``.
    :type atom_df_processing_funcs: List[Callable], optional
    :param hetatom_df_processing_funcs: List of functions to process the hetatom
        dataframe. These must take in a DataFrame and return a DataFrame.
    :type hetatom_df_processing_funcs: List[Callable], optional
    :param granularity: The level of granularity for the graph. This determines
        the node definition. Acceptable values include: ``"centroids"``,
        ``"atoms"``, any of the atom_names in the PDB file (e.g. ``"CA"``,
        ``"CB"``, ``"OG"``, etc.).
        See: :const:`~graphein.protein.config.GRAPH_ATOMS` and
        :const:`~graphein.protein.config.GRANULARITY_OPTS`.
    :type granularity: str
    :param insertions: Whether or not to keep insertions. Defaults to ``False``.
    :param insertions: bool
    :param alt_locs: Whether or not to keep alternatively located atoms.
    :param alt_locs: bool
    :param deprotonate: Whether or not to remove hydrogen atoms (i.e.
        deprotonation).
    :type deprotonate: bool
    :param keep_hets: Hetatoms to keep. Defaults to an empty list (``[]``).
        To keep a hetatom, pass it inside a list of hetatom names to keep.
    :type keep_hets: List[str]
    :param chain_selection: Which protein chain to select. Defaults to
        ``"all"``. Eg can use ``"ACF"`` to select 3 chains (``A``, ``C`` &
        ``F``)
    :type chain_selection: str
    :return: A protein dataframe that can be consumed by other graph
        construction functions.
    :rtype: pd.DataFrame
    """
    protein_df = label_node_id(
        protein_df, granularity=granularity, insertions=insertions
    )
    # TODO: Need to properly define what "granularity" is supposed to do.
    atoms = filter_dataframe(
        protein_df,
        by_column="record_name",
        list_of_values=["ATOM"],
        boolean=True,
    )
    hetatms = filter_dataframe(
        protein_df,
        by_column="record_name",
        list_of_values=["HETATM"],
        boolean=True,
    )

    # This block enables processing via a list of supplied functions operating
    # on the atom and hetatom DataFrames. If these are provided, the DataFrame
    # returned will be computed only from these and the default workflow
    # below this block will not execute.
    if atom_df_processing_funcs is not None:
        for func in atom_df_processing_funcs:
            atoms = func(atoms)
        if hetatom_df_processing_funcs is None:
            return atoms

    if hetatom_df_processing_funcs is not None:
        for func in hetatom_df_processing_funcs:
            hetatms = func(hetatms)
        return pd.concat([atoms, hetatms])

    if keep_hets:
        hetatms_to_keep = filter_hetatms(hetatms, keep_hets)
        atoms = pd.concat([atoms] + hetatms_to_keep)

    # Deprotonate structure by removing H atoms
    if deprotonate:
        atoms = deprotonate_structure(atoms)

    # Restrict DF to desired granularity
    if granularity == "atom":
        pass
    elif granularity in {"centroids", "rna_centroid"}:
        atoms = convert_structure_to_centroids(atoms)
    elif granularity == "rna_atom":
        atoms = subset_structure_to_rna(atoms)
    else:
        atoms = subset_structure_to_atom_type(atoms, granularity)

    protein_df = atoms

    # Remove alt_loc residues
    if alt_locs != "include":
        protein_df = remove_alt_locs(protein_df, keep=alt_locs)

    # Remove inserted residues
    if not insertions:
        protein_df = remove_insertions(protein_df)

    # perform chain selection
    protein_df = select_chains(protein_df, chain_selection=chain_selection)

    log.debug(f"Detected {len(protein_df)} total nodes")

    # Sort dataframe to place HETATMs and reset index
    protein_df = sort_dataframe(protein_df)
    protein_df = protein_df.reset_index(drop=True)

    return protein_df


def sort_dataframe(df: pd.DataFrame) -> pd.DataFrame:
    """Sorts a protein dataframe by chain->residue number->atom number

    This is useful for distributing hetatms/modified residues through the DF.

    :param df: Protein dataframe to sort.
    :type df: pd.DataFrame
    :return: Sorted protein dataframe.
    :rtype: pd.DataFrame
    """
    return df.sort_values(
        by=["chain_id", "residue_number", "atom_number", "insertion"]
    )


def select_chains(
    protein_df: pd.DataFrame,
    chain_selection: Union[str, List[str]],
) -> pd.DataFrame:
    """
    Extracts relevant chains from ``protein_df``.

    :param protein_df: pandas DataFrame of PDB subsetted to relevant atoms
        (``CA``, ``CB``).
    :type protein_df: pd.DataFrame
    :param chain_selection: Specifies chains that should be extracted from
        the larger complexed structure. If chain_selection is ``"all"``, all
        chains will be selected. Otherwise, provide a list of strings.
    :type chain_selection: Union[str, List[str]]
    :return: Protein structure dataframe containing only entries in the
        chain selection.
    :rtype: pd.DataFrame
    """
    if chain_selection != "all":
        if isinstance(chain_selection, str):
            raise ValueError(
                "Only 'all' is a valid string for chain selection. Otherwise use a list of strings: e.g. ['A', 'B', 'C']"
            )
        protein_df = filter_dataframe(
            protein_df,
            by_column="chain_id",
            list_of_values=chain_selection,
            boolean=True,
        )

    return protein_df


def initialise_graph_with_metadata(
    protein_df: pd.DataFrame,
    raw_pdb_df: pd.DataFrame,
    granularity: str,
    name: Optional[str] = None,
    pdb_code: Optional[str] = None,
    path: Optional[str] = None,
) -> nx.Graph:
    """
    Initializes the nx Graph object with initial metadata.

    :param protein_df: Processed DataFrame of protein structure.
    :type protein_df: pd.DataFrame
    :param raw_pdb_df: Unprocessed dataframe of protein structure for comparison
        and traceability downstream.
    :type raw_pdb_df: pd.DataFrame
    :param granularity: Granularity of the graph (eg ``"atom"``, ``"CA"``,
        ``"CB"`` etc or ``"centroid"``). See:
        :const:`~graphein.protein.config.GRAPH_ATOMS` and
        :const:`~graphein.protein.config.GRANULARITY_OPTS`.
    :type granularity: str
    :param name: specified given name for the graph. If None, the PDB code or
        the file name will be used to name the graph.
    :type name: Optional[str], defaults to ``None``
    :param pdb_code: PDB ID / Accession code, if the PDB is available on the
        PDB database.
    :type pdb_code: Optional[str], defaults to ``None``.
    :param path: path to local PDB or MMTF file, if constructing a graph from a
        local file.
    :type path: Optional[str], defaults to ``None``.
    :return: Returns initial protein structure graph with metadata.
    :rtype: nx.Graph
    """
    if path is not None and isinstance(path, Path):
        path = os.fsdecode(path)

    # Get name for graph if no name was provided
    if name is None:
        if path is not None:
            name = get_protein_name_from_filename(path)
        else:
            name = pdb_code

    G = nx.Graph(
        name=name,
        pdb_code=pdb_code,
        path=path,
        chain_ids=list(protein_df["chain_id"].unique()),
        pdb_df=protein_df,
        raw_pdb_df=raw_pdb_df,
        rgroup_df=compute_rgroup_dataframe(raw_pdb_df),
        coords=np.asarray(protein_df[["x_coord", "y_coord", "z_coord"]]),
    )

    # Create graph and assign intrinsic graph-level metadata
    G.graph["node_type"] = granularity

    # Add Sequences to graph metadata
    for c in G.graph["chain_ids"]:
        if granularity == "rna_atom":
            sequence = protein_df.loc[protein_df["chain_id"] == c][
                "residue_name"
            ].str.cat()
        elif granularity == "atom":
            sequence = (
                protein_df.loc[
                    (protein_df["chain_id"] == c)
                    & (protein_df["atom_name"] == "CA")
                ]["residue_name"]
                .apply(three_to_one_with_mods)
                .str.cat()
            )
        else:
            sequence = (
                protein_df.loc[protein_df["chain_id"] == c]["residue_name"]
                .apply(three_to_one_with_mods)
                .str.cat()
            )
        G.graph[f"sequence_{c}"] = sequence
    return G


def add_nodes_to_graph(
    G: nx.Graph,
    protein_df: Optional[pd.DataFrame] = None,
    verbose: bool = False,
) -> nx.Graph:
    """Add nodes into protein graph.

    :param G: ``nx.Graph`` with metadata to populate with nodes.
    :type G: nx.Graph
    :param protein_df: DataFrame of protein structure containing nodes & initial
        node metadata to add to the graph. Defaults to ``None``.
    :type protein_df: pd.DataFrame, optional
    :param verbose: Controls verbosity of this step. Defaults to ``False``.
    :type verbose: bool
    :returns: nx.Graph with nodes added.
    :rtype: nx.Graph
    """

    # If no protein dataframe is supplied, use the one stored in the Graph
    # object
    if protein_df is None:
        protein_df: pd.DataFrame = G.graph["pdb_df"]
    # Assign intrinsic node attributes
    chain_id = protein_df["chain_id"].apply(str)
    residue_name = protein_df["residue_name"]
    residue_number = protein_df["residue_number"]  # .apply(str)
    coords = np.asarray(protein_df[["x_coord", "y_coord", "z_coord"]])
    b_factor = protein_df["b_factor"]
    atom_type = protein_df["atom_name"]
    nodes = protein_df["node_id"]
    element_symbol = protein_df["element_symbol"]
    G.add_nodes_from(nodes)

    # Set intrinsic node attributes
    nx.set_node_attributes(G, dict(zip(nodes, chain_id)), "chain_id")
    nx.set_node_attributes(G, dict(zip(nodes, residue_name)), "residue_name")
    nx.set_node_attributes(
        G, dict(zip(nodes, residue_number)), "residue_number"
    )
    nx.set_node_attributes(G, dict(zip(nodes, atom_type)), "atom_type")
    nx.set_node_attributes(
        G, dict(zip(nodes, element_symbol)), "element_symbol"
    )
    nx.set_node_attributes(G, dict(zip(nodes, coords)), "coords")
    nx.set_node_attributes(G, dict(zip(nodes, b_factor)), "b_factor")

    # TODO: include charge, line_idx for traceability?
    if verbose:
        print(nx.info(G))
        print(G.nodes())

    return G


def calculate_centroid_positions(
    atoms: pd.DataFrame, verbose: bool = False
) -> pd.DataFrame:
    """
    Calculates position of sidechain centroids.

    :param atoms: ATOM df of protein structure.
    :type atoms: pd.DataFrame
    :param verbose: bool controlling verbosity.
    :type verbose: bool
    :return: centroids (df).
    :rtype: pd.DataFrame
    """
    centroids = (
        atoms.groupby(
            ["residue_number", "chain_id", "residue_name", "insertion"]
        )
        .mean(numeric_only=True)[["x_coord", "y_coord", "z_coord"]]
        .reset_index()
    )
    if verbose:
        print(f"Calculated {len(centroids)} centroid nodes")
    log.debug(f"Calculated {len(centroids)} centroid nodes")
    return centroids


def compute_edges(
    G: nx.Graph,
    funcs: List[Callable],
    get_contacts_config: Optional[GetContactsConfig] = None,
) -> nx.Graph:
    """
    Computes edges for the protein structure graph. Will compute a pairwise
    distance matrix between nodes which is
    added to the graph metadata to facilitate some edge computations.

    :param G: nx.Graph with nodes to add edges to.
    :type G: nx.Graph
    :param funcs: List of edge construction functions.
    :type funcs: List[Callable]
    :param get_contacts_config: Config object for ``GetContacts`` if
        intramolecular edges are being used.
    :type get_contacts_config: graphein.protein.config.GetContactsConfig
    :return: Graph with added edges.
    :rtype: nx.Graph
    """
    # This control flow prevents unnecessary computation of the distance
    # matrices
    if "config" in G.graph:
        if G.graph["config"].granularity == "atom":
            G.graph["atomic_dist_mat"] = compute_distmat(G.graph["pdb_df"])
        else:
            G.graph["dist_mat"] = compute_distmat(G.graph["pdb_df"])

    for func in funcs:
        func(G)

    return add_distance_to_edges(G)


def construct_graph(
    config: Optional[ProteinGraphConfig] = None,
    name: Optional[str] = None,
    path: Optional[Union[str, os.PathLike]] = None,
    uniprot_id: Optional[str] = None,
    pdb_code: Optional[str] = None,
    df: Optional[pd.DataFrame] = None,
    chain_selection: Union[str, List[str]] = "all",
    model_index: int = 1,
    df_processing_funcs: Optional[List[Callable]] = None,
    edge_construction_funcs: Optional[List[Callable]] = None,
    edge_annotation_funcs: Optional[List[Callable]] = None,
    node_annotation_funcs: Optional[List[Callable]] = None,
    graph_annotation_funcs: Optional[List[Callable]] = None,
    verbose: bool = True,
) -> nx.Graph:
    """
    Constructs protein structure graph from a ``pdb_code``, ``path``,
    ``uniprot_id`` or a BioPandas DataFrame containing ``ATOM`` data.

    Users can provide a :class:`~graphein.protein.config.ProteinGraphConfig`
    object to specify construction parameters.

    However, config parameters can be overridden by passing arguments directly
    to the function.

    :param config: :class:`~graphein.protein.config.ProteinGraphConfig` object.
        If ``None``, defaults to config in ``graphein.protein.config``.
    :type config: graphein.protein.config.ProteinGraphConfig, optional
    :param name: an optional given name for the graph. the PDB ID or PDB file
        name will be used if not specified.
    :type name: str, optional
    :param path: Path to PDB or MMTF file when constructing a graph from a
        local pdb file. Default is ``None``.
    :type path: Optional[str], defaults to ``None``
    :param pdb_code: A 4-character PDB ID / accession to be used to construct
        the graph, if available. Default is ``None``.
    :type pdb_code: Optional[str], defaults to ``None``
    :param uniprot_id: UniProt accession ID to build graph from AlphaFold2DB.
        Default is ``None``.
    :type uniprot_id: str, optional
    :param df: Pandas dataframe containing ATOM data to build graph from.
        Default is ``None``.
    :type df: pd.DataFrame, optional
    :param chain_selection: List of strings denoting polypeptide chains to
        include in graph. E.g ``["A", "B", "D", "F"]`` or ``"all"``. Default is ``"all"``.
    :type chain_selection: str
    :param model_index: Index of model to use in the case of structural
        ensembles. Default is ``1``.
    :type model_index: int
    :param df_processing_funcs: List of dataframe processing functions.
        Default is ``None``.
    :type df_processing_funcs: List[Callable], optional
    :param edge_construction_funcs: List of edge construction functions.
        Default is ``None``.
    :type edge_construction_funcs: List[Callable], optional
    :param edge_annotation_funcs: List of edge annotation functions.
        Default is ``None``.
    :type edge_annotation_funcs: List[Callable], optional
    :param node_annotation_funcs: List of node annotation functions.
        Default is ``None``.
    :type node_annotation_funcs: List[Callable], optional
    :param graph_annotation_funcs: List of graph annotation function.
        Default is ``None``.
    :type graph_annotation_funcs: List[Callable]
    :param verbose: Controls the verbosity.
        Default is ``True``.
    :type verbose: bool
    :return: Protein Structure Graph
    :rtype: nx.Graph
    """

    if pdb_code is None and path is None and uniprot_id is None and df is None:
        raise ValueError(
            "Either a PDB ID, UniProt ID, a dataframe or a path to a local PDB file"
            " must be specified to construct a graph"
        )
    if path is not None and isinstance(path, Path):
        path = os.fsdecode(path)

    # If no config is provided, use default
    if config is None:
        config = ProteinGraphConfig()

    # Use progress tracking context if in verbose mode
    context = Progress(transient=True) if verbose else nullcontext()
    with context as progress:
        if verbose:
            task1 = progress.add_task("Reading PDB file...", total=1)
            progress.advance(task1)

        # If config params are provided, overwrite them
        config.protein_df_processing_functions = (
            df_processing_funcs
            if config.protein_df_processing_functions is None
            else config.protein_df_processing_functions
        )
        config.edge_construction_functions = (
            edge_construction_funcs
            if config.edge_construction_functions is None
            else config.edge_construction_functions
        )
        config.node_metadata_functions = (
            node_annotation_funcs
            if config.node_metadata_functions is None
            else config.node_metadata_functions
        )
        config.graph_metadata_functions = (
            graph_annotation_funcs
            if config.graph_metadata_functions is None
            else config.graph_metadata_functions
        )
        config.edge_metadata_functions = (
            edge_annotation_funcs
            if config.edge_metadata_functions is None
            else config.edge_metadata_functions
        )
        if df is None:
            raw_df = read_pdb_to_dataframe(
                path,
                pdb_code,
                uniprot_id,
                model_index=model_index,
            )
        else:
            raw_df = df

        if verbose:
            task2 = progress.add_task("Processing PDB dataframe...", total=1)
        raw_df = sort_dataframe(raw_df)
        protein_df = process_dataframe(
            raw_df,
            chain_selection=chain_selection,
            granularity=config.granularity,
            insertions=config.insertions,
            alt_locs=config.alt_locs,
            keep_hets=config.keep_hets,
            atom_df_processing_funcs=config.protein_df_processing_functions,
            hetatom_df_processing_funcs=config.protein_df_processing_functions,
            deprotonate=config.deprotonate,
        )

        if verbose:
            progress.advance(task2)

            task3 = progress.add_task("Initializing graph...", total=1)
        # Initialise graph with metadata
        g = initialise_graph_with_metadata(
            protein_df=protein_df,
            raw_pdb_df=raw_df,
            name=name,
            pdb_code=pdb_code,
            path=path,
            granularity=config.granularity,
        )
        # Add nodes to graph
        g = add_nodes_to_graph(g)

        # Add config to graph
        g.graph["config"] = config

        # Annotate additional node metadata
        if config.node_metadata_functions is not None:
            g = annotate_node_metadata(g, config.node_metadata_functions)

        if verbose:
            progress.advance(task3)
            task4 = progress.add_task("Constructing edges...", total=1)
        # Compute graph edges
        g = compute_edges(
            g,
            funcs=config.edge_construction_functions,
            get_contacts_config=None,
        )

        if verbose:
            progress.advance(task4)

    # Annotate additional graph metadata
    if config.graph_metadata_functions is not None:
        g = annotate_graph_metadata(g, config.graph_metadata_functions)

    # Annotate additional edge metadata
    if config.edge_metadata_functions is not None:
        g = annotate_edge_metadata(g, config.edge_metadata_functions)

    return g


def _mp_graph_constructor(
    args: Tuple[str, str, int], source: str, config: ProteinGraphConfig
) -> Union[nx.Graph, None]:
    """
    Protein graph constructor for use in multiprocessing several protein
    structure graphs.

    :param args: Tuple of pdb code/path and the chain selection for that PDB.
    :type args: Tuple[str, str]
    :param use_pdb_code: Whether we are using ``"pdb_code"``s, ``path``s
        (to PDB or MMTF files) or ``"uniprot_id"``s.
    :type use_pdb_code: bool
    :param config: Protein structure graph construction config
        (see: :class:`graphein.protein.config.ProteinGraphConfig`).
    :type config: ProteinGraphConfig
    :return: Protein structure graph or ``None`` if an error is encountered.
    :rtype: Union[nx.Graph, None]
    """
    log.info(f"Constructing graph for: {args[0]}. Chain selection: {args[1]}. \
            Model index: {args[2]}")
    func = partial(construct_graph, config=config)
    try:
        if source == "pdb_code":
            return func(
                pdb_code=args[0], chain_selection=args[1], model_index=args[2]
            )
        elif source == "path":
            return func(
                path=args[0], chain_selection=args[1], model_index=args[2]
            )
        elif source == "uniprot_id":
            return func(
                uniprot_id=args[0],
                chain_selection=args[1],
                model_index=args[2],
            )

    except Exception as ex:
        log.info(f"Graph construction error (PDB={args[0]})! \
                {traceback.format_exc()}")
        log.info(ex)
        return None


def construct_graphs_mp(
    pdb_code_it: Optional[List[str]] = None,
    path_it: Optional[List[str]] = None,
    uniprot_id_it: Optional[List[str]] = None,
    chain_selections: Optional[Union[List[List[str]], List[str]]] = None,
    model_indices: Optional[List[str]] = None,
    config: ProteinGraphConfig = ProteinGraphConfig(),
    num_cores: int = 16,
    return_dict: bool = True,
    out_path: Optional[str] = None,
) -> Union[List[nx.Graph], Dict[str, nx.Graph]]:
    """
    Constructs protein graphs for a list of pdb codes or pdb paths using
    multiprocessing.

    :param pdb_code_it: List of pdb codes to use for protein graph construction
    :type pdb_code_it: Optional[List[str]], defaults to ``None``
    :param path_it: List of paths to PDB or MMTF files to use for protein graph
        construction.
    :type path_it: Optional[List[str]], defaults to ``None``
    :param chain_selections: List of chains to select from the protein
        structures (e.g. ``[["A", "B" "C"], ["A"], ["L"], ["C", "D"]...]``).
    :type chain_selections: Optional[List[str]], defaults to ``None``
    :param model_indices: List of model indices to use for protein graph
        construction. Only relevant for structures containing ensembles of
        models.
    :type model_indices: Optional[List[str]], defaults to ``None``
    :param config: ProteinGraphConfig to use.
    :type config: graphein.protein.config.ProteinGraphConfig, defaults to
        default config params.
    :param num_cores: Number of cores to use for multiprocessing. The more the
        merrier.
    :type num_cores: int, defaults to ``16``
    :param return_dict: Whether or not to return a dictionary
        (indexed by pdb codes/paths) or a list of graphs.
    :type return_dict: bool, default to ``True``
    :param out_path: Path to save the graphs to. If ``None``, graphs are not
        saved to disk.
    :type out_path: Optional[str], defaults to ``None``
    :return: Iterable of protein graphs. ``None`` values indicate there was a
        problem in constructing the graph for this particular pdb.
    :rtype: Union[List[nx.Graph], Dict[str, nx.Graph]]
    """
    assert (
        pdb_code_it is not None or path_it is not None
    ), "Iterable of pdb codes, pdb paths or uniprot IDs required."

    if pdb_code_it is not None:
        pdbs = pdb_code_it
        source = "pdb_code"

    if path_it is not None:
        pdbs = path_it
        source = "path"

    if uniprot_id_it is not None:
        pdbs = uniprot_id_it
        source = "uniprot_id"

    if chain_selections is None:
        chain_selections = ["all"] * len(pdbs)

    if model_indices is None:
        model_indices = [1] * len(pdbs)

    constructor = partial(_mp_graph_constructor, source=source, config=config)

    graphs = list(
        process_map(
            constructor,
            [
                (pdb, chain_selections[i], model_indices[i])
                for i, pdb in enumerate(pdbs)
            ],
            max_workers=num_cores,
        )
    )
    if out_path is not None:
        [
            nx.write_gpickle(
                g, str(f"{out_path}/" + f"{g.graph['name']}.pickle")
            )
            for g in graphs
        ]

    if return_dict:
        graphs = {pdb: graphs[i] for i, pdb in enumerate(pdbs)}

    return graphs


def compute_chain_graph(
    g: nx.Graph,
    chain_list: Optional[List[str]] = None,
    remove_self_loops: bool = False,
    return_weighted_graph: bool = False,
) -> Union[nx.Graph, nx.MultiGraph]:
    """Computes a chain-level graph from a protein structure graph.

    This graph features nodes as individual chains in a complex and edges as
    the interactions between constituent nodes in each chain. You have the
    option of returning an unweighted graph (multigraph,
    ``return_weighted_graph=False``) or a weighted graph
    (``return_weighted_graph=True``). The difference between these is the
    unweighted graph features and edge for each interaction between chains
    (ie the number of edges will be equal to the number of edges in the input
    protein structure graph), while the weighted graph sums these interactions
    to a single edge between chains with the counts stored as features.

    :param g: A protein structure graph to compute the chain graph of.
    :type g: nx.Graph
    :param chain_list: A list of chains to extract from the input graph.
        If ``None``, all chains will be used. This is provided as input to
        ``extract_subgraph_from_chains``. Default is ``None``.
    :type chain_list: Optional[List[str]]
    :param remove_self_loops: Whether to remove self-loops from the graph.
        Default is False.
    :type remove_self_loops: bool
    :return: A chain-level graph.
    :rtype: Union[nx.Graph, nx.MultiGraph]
    """
    # If we are extracting specific chains, do it here.
    if chain_list is not None:
        g = extract_subgraph_from_chains(g, chain_list)

    # Initialise new graph with Metadata
    h = nx.MultiGraph()
    h.graph = g.graph
    h.graph["node_type"] = "chain"

    # Set nodes
    nodes_per_chain = {chain: 0 for chain in g.graph["chain_ids"]}
    sequences = {chain: "" for chain in g.graph["chain_ids"]}
    for n, d in g.nodes(data=True):
        nodes_per_chain[d["chain_id"]] += 1
        sequences[d["chain_id"]] += RESI_THREE_TO_1[d["residue_name"]]

    h.add_nodes_from(g.graph["chain_ids"])

    for n, d in h.nodes(data=True):
        d["num_residues"] = nodes_per_chain[n]
        d["sequence"] = sequences[n]

    # Add edges
    for u, v, d in g.edges(data=True):
        h.add_edge(
            g.nodes[u]["chain_id"], g.nodes[v]["chain_id"], kind=d["kind"]
        )
    # Remove self-loops if necessary. Checks for equality between nodes in a
    # given edge.
    if remove_self_loops:
        edges_to_remove: List[Tuple[str]] = [
            (u, v) for u, v in h.edges() if u == v
        ]
        h.remove_edges_from(edges_to_remove)

    # Compute a weighted graph if required.
    if return_weighted_graph:
        return compute_weighted_graph_from_multigraph(h)
    return h


def compute_weighted_graph_from_multigraph(g: nx.MultiGraph) -> nx.Graph:
    """Computes a weighted graph from a multigraph.

    This function is used to convert a multigraph to a weighted graph. The
    weights of the edges are the number of interactions between the nodes.

    :param g: A multigraph.
    :type g: nx.MultiGraph
    :return: A weighted graph.
    :rtype: nx.Graph
    """
    H = nx.Graph()
    H.graph = g.graph
    H.add_nodes_from(g.nodes(data=True))
    for u, v, d in g.edges(data=True):
        if H.has_edge(u, v):
            H[u][v]["weight"] += len(d["kind"])
            H[u][v]["kind"].update(d["kind"])
            for kind in list(d["kind"]):
                try:
                    H[u][v][kind] += 1
                except KeyError:
                    H[u][v][kind] = 1
        else:
            H.add_edge(u, v, weight=len(d["kind"]), kind=d["kind"])
            for kind in list(d["kind"]):
                H[u][v][kind] = 1
    return H


def number_groups_of_runs(list_of_values: List[Any]) -> List[str]:
    """Numbers groups of runs in a list of values.

    E.g. ``["A", "A", "B", "A", "A", "A", "B", "B"] ->
    ["A1", "A1", "B1", "A2", "A2", "A2", "B2", "B2"]``

    :param list_of_values: List of values to number.
    :type list_of_values: List[Any]
    :return: List of numbered values.
    :rtype: List[str]
    """
    df = pd.DataFrame({"val": list_of_values})
    df["idx"] = df["val"].shift() != df["val"]
    df["sum"] = df.groupby("val")["idx"].cumsum(numeric_only=True)
    return list(df["val"].astype(str) + df["sum"].astype(str))


def compute_secondary_structure_graph(
    g: nx.Graph,
    allowable_ss_elements: Optional[List[str]] = None,
    remove_non_ss: bool = True,
    remove_self_loops: bool = False,
    return_weighted_graph: bool = False,
) -> Union[nx.Graph, nx.MultiGraph]:
    """Computes a secondary structure graph from a protein structure graph.

    :param g: A protein structure graph to compute the secondary structure
        graph of.
    :type g: nx.Graph
    :param remove_non_ss: Whether to remove non-secondary structure nodes from
        the graph. These are denoted as ``"-"`` by DSSP. Default is True.
    :type remove_non_ss: bool
    :param remove_self_loops: Whether to remove self-loops from the graph.
        Default is ``False``.
    :type remove_self_loops: bool
    :param return_weighted_graph: Whether to return a weighted graph.
        Default is False.
    :type return_weighted_graph: bool
    :raises ProteinGraphConfigurationError: If the protein structure graph is
        not configured correctly with secondary structure assignments on all
        nodes.
    :return: A secondary structure graph.
    :rtype: Union[nx.Graph, nx.MultiGraph]
    """
    # Initialise list of secondary structure elements we use to build the graph
    ss_list: List[str] = []

    # Check nodes have secondary structure assignment & store them in list
    for _, d in g.nodes(data=True):
        if "ss" not in d.keys():
            raise ProteinGraphConfigurationError(
                "Secondary structure not defined for all nodes."
            )
        ss_list.append(d["ss"])

    # Number SS elements
    ss_list: pd.Series = pd.Series(number_groups_of_runs(ss_list))
    ss_list.index = list(g.nodes())

    # Remove unstructured elements if necessary
    if remove_non_ss:
        ss_list = ss_list[~ss_list.str.contains("-")]
    # Subset to only allowable SS elements if necessary
    if allowable_ss_elements:
        ss_list = ss_list[
            ss_list.str.contains("|".join(allowable_ss_elements))
        ]

    constituent_residues: Dict[str, List[str]] = ss_list.index.groupby(
        ss_list.values
    )
    constituent_residues = {
        k: list(v) for k, v in constituent_residues.items()
    }
    residue_counts: Dict[str, int] = ss_list.groupby(ss_list).count().to_dict()

    # Add Nodes from secondary structure list
    h = nx.MultiGraph()
    h.add_nodes_from(ss_list)
    nx.set_node_attributes(h, residue_counts, "residue_counts")
    nx.set_node_attributes(h, constituent_residues, "constituent_residues")
    # Assign ss
    for n, d in h.nodes(data=True):
        d["ss"] = n[0]

    # Add graph-level metadata
    h.graph = g.graph
    h.graph["node_type"] = "secondary_structure"

    # Iterate over edges in source graph and add SS-SS edges to new graph.
    for u, v, d in g.edges(data=True):
        try:
            h.add_edge(
                ss_list[u], ss_list[v], kind=d["kind"], source=f"{u}_{v}"
            )
        except KeyError as e:
            log.debug(f"Edge {u}-{v} not added to secondary structure graph. \
                Reason: {e} not in graph")

    # Remove self-loops if necessary.
    # Checks for equality between nodes in a given edge.
    if remove_self_loops:
        edges_to_remove: List[Tuple[str]] = [
            (u, v) for u, v in h.edges() if u == v
        ]
        h.remove_edges_from(edges_to_remove)

    # Create weighted graph from h
    if return_weighted_graph:
        return compute_weighted_graph_from_multigraph(h)
    return h


def compute_line_graph(g: nx.Graph, repopulate_data: bool = True) -> nx.Graph:
    """Computes the line graph of a graph.

    The line graph of a graph G has a node for each edge in G and an edge
    joining those nodes if the two edges in G share a common node. For directed
    graphs, nodes are adjacent exactly when the edges they represent form a
    directed path of length two.

    The nodes of the line graph are 2-tuples of nodes in the original graph (or
    3-tuples for multigraphs, with the key of the edge as the third element).

    :param g: Graph to compute the line graph of.
    :type g: nx.Graph
    :param repopulate_data: Whether or not to map node and edge data to edges
        and nodes of the line graph, defaults to True
    :type repopulate_data: bool, optional
    :return: Line graph of g.
    :rtype: nx.Graph
    """
    l_g = nx.generators.line_graph(g)
    l_g.graph = g.graph

    if repopulate_data:
        source_edge_data = {(u, v): d for u, v, d in g.edges(data=True)}
        nx.set_node_attributes(l_g, source_edge_data)

        node_list = {}
        for u, v, d in l_g.edges(data=True):
            node_union = u + v
            for n in node_union:
                if node_union.count(n) > 1:
                    node_list[(u, v)] = n
                    break

        source_node_data = {k: g.nodes[v] for k, v in node_list.items()}
        nx.set_edge_attributes(l_g, source_node_data)
    return l_g



"""Utilities for computing edges."""

from typing import Optional

from loguru import logger as log

from graphein.utils.dependencies import import_message

from .types import CoordTensor, EdgeTensor, OptTensor

try:
    import torch
except ImportError:
    message = import_message(
        "graphein.protein.tensor.edges",
        "torch",
        conda_channel="pytorch",
        pip_install=True,
    )
    log.warning(message)

try:
    from torch_geometric.nn.pool import knn_graph, radius_graph
except ImportError:
    message = import_message(
        "graphein.protein.tensor.edges",
        "torch_geometric",
        "pyg",
        pip_install=True,
    )


def compute_edges(
    x: torch.Tensor,
    edge_type: str = "knn_8",
    batch: OptTensor = None,
    **kwargs,
) -> EdgeTensor:
    """Computes edges for a given node feature (or coordinate) matrix ``x``.

    .. code-block:: python
        import torch
        import graphein.protein.tensor as gpt

        x = torch.rand(10, 3)

        gpt.compute_edges(x, "knn_8") # KNN with k = 8
        gpt.compute_edges(x, "eps_8") # Radius graph with r = 8

    :param x: Node features or positions
    :param edge_type: Str denoting type of edges in form ``{edgetype}_{value}``.
        E.g. ``"knn_8"`` for KNN with ``k=8``,
        ``"eps_6"`` for radius graph with ``r=6``.
    :returns: Edge indices
    :rtype: EdgeTensor
    """

    edge_type = edge_type.lower()
    etype = edge_type.split("_")[0]
    val = edge_type.split("_")[1]

    if etype == "knn":
        return knn_edges(x=x, k=int(val), batch=batch, **kwargs)
    elif etype == "eps":
        return radius_edges(x=x, radius=float(val), batch=batch, **kwargs)
    else:
        raise ValueError(f"Edge type {edge_type} not recognised.")


def radius_edges(
    x: torch.Tensor,
    radius: float = 8.0,
    batch: Optional[torch.Tensor] = None,
    loop: bool = False,
    max_num_neighbors: int = 32,
    flow: str = "source_to_target",
    num_workers: int = 1,
) -> EdgeTensor:
    r"""Computes graph edges to all points within a given distance.

    .. code-block:: python

            import torch
            import graphein.protein.tensor

            x = torch.Tensor([[-1, -1], [-1, 1], [1, -1], [1, 1]])
            batch = torch.tensor([0, 0, 0, 0])
            edge_index = gpt.radius_graph(x, r=1.5, batch=batch, loop=False)

    Args:
        x (Tensor): Node feature matrix
            :math:`\mathbf{X} \in \mathbb{R}^{N \times F}`.
        r (float): The radius.
        batch (LongTensor, optional): Batch vector
            :math:`\mathbf{b} \in {\{ 0, \ldots, B-1\}}^N`, which assigns each
            node to a specific example. (default: :obj:`None`)
        loop (bool, optional): If :obj:`True`, the graph will contain
            self-loops. (default: :obj:`False`)
        max_num_neighbors (int, optional): The maximum number of neighbors to
            return for each element in :obj:`y`. (default: :obj:`32`)
        flow (string, optional): The flow direction when using in combination
            with message passing (:obj:`"source_to_target"` or
            :obj:`"target_to_source"`). (default: :obj:`"source_to_target"`)
        num_workers (int): Number of workers to use for computation. Has no
            effect in case :obj:`batch` is not :obj:`None`, or the input lies
            on the GPU. (default: :obj:`1`)

    """
    return radius_graph(
        x,
        r=radius,
        batch=batch,
        loop=loop,
        max_num_neighbors=max_num_neighbors,
        flow=flow,
        num_workers=num_workers,
    )


def knn_edges(
    x: torch.Tensor,
    k: int = 10,
    batch: Optional[torch.Tensor] = None,
    loop: bool = False,
    flow: str = "source_to_target",
    num_workers: int = 1,
) -> EdgeTensor:
    r"""Computes graph edges to the nearest :obj:`k` points.

    .. code-block:: python

        import torch
        import graphein.protein.tensor as gpt

        x = torch.Tensor([[-1, -1], [-1, 1], [1, -1], [1, 1]])
        batch = torch.tensor([0, 0, 0, 0])
        edge_index = gpt.knn_graph(x, k=2, batch=batch, loop=False)

    Args:
        x (Tensor): Node feature matrix
            :math:`\mathbf{X} \in \mathbb{R}^{N \times F}`.
        k (int): The number of neighbors.
        batch (LongTensor, optional): Batch vector
            :math:`\mathbf{b} \in {\{ 0, \ldots, B-1\}}^N`, which assigns each
            node to a specific example. (default: :obj:`None`)
        loop (bool, optional): If :obj:`True`, the graph will contain
            self-loops. (default: :obj:`False`)
        flow (string, optional): The flow direction when using in combination
            with message passing (:obj:`"source_to_target"` or
            :obj:`"target_to_source"`). (default: :obj:`"source_to_target"`)
        cosine (boolean, optional): If :obj:`True`, will use the cosine
            distance instead of euclidean distance to find nearest neighbors.
            (default: :obj:`False`)
        num_workers (int): Number of workers to use for computation. Has no
            effect in case :obj:`batch` is not :obj:`None`, or the input lies
            on the GPU. (default: :obj:`1`)
    """
    return knn_graph(
        x,
        k=k,
        batch=batch,
        loop=loop,
        flow=flow,
        num_workers=num_workers,
    )


def edge_distances(x: CoordTensor, edge_index: EdgeTensor, p: float = 2):
    """Computes the distances between edges

    :param x: Node positions
    :param edge_index: Edge indices
    :param p: The norm degree. Can be negative. Default: ``2``.``
    :returns: Edge distances
    :rtype: torch.Tensor
    """
    return torch.pairwise_distance(x[edge_index[0]], x[edge_index[1]], p=p)


def contact_map(x: CoordTensor) -> torch.Tensor:
    """Computes the contact map for a set of coordinates

    :param x: Node positions
    """
    return torch.cdist(x, x)



"""Functions for computing biochemical edges of graphs."""

# Graphein
# Author: Eric Ma, Arian Jamasb <arian@jamasb.io>
# License: MIT
# Project Website: https://github.com/a-r-j/graphein
# Code Repository: https://github.com/a-r-j/graphein
from __future__ import annotations

import itertools
from itertools import combinations, product
from typing import Dict, Iterable, List, Optional, Tuple, Union

import networkx as nx
import numpy as np
import pandas as pd
from loguru import logger as log
from scipy.spatial import Delaunay
from scipy.spatial.distance import pdist, squareform
from sklearn.neighbors import NearestNeighbors, kneighbors_graph

from graphein.protein.resi_atoms import (
    AA_RING_ATOMS,
    AROMATIC_RESIS,
    BACKBONE_ATOMS,
    BOND_TYPES,
    CATION_PI_RESIS,
    CATION_RESIS,
    DISULFIDE_ATOMS,
    DISULFIDE_RESIS,
    HYDROPHOBIC_RESIS,
    IONIC_RESIS,
    NEG_AA,
    PI_RESIS,
    POS_AA,
    RING_NORMAL_ATOMS,
    SALT_BRIDGE_ANIONS,
    SALT_BRIDGE_ATOMS,
    SALT_BRIDGE_CATIONS,
    SALT_BRIDGE_RESIDUES,
    SULPHUR_RESIS,
    VDW_RADII,
)
from graphein.protein.utils import filter_dataframe

INFINITE_DIST = 10_000.0  # np.inf leads to errors in some cases


def compute_distmat(pdb_df: pd.DataFrame) -> pd.DataFrame:
    """
    Compute pairwise Euclidean distances between every atom.

    Design choice: passed in a ``pd.DataFrame`` to enable easier testing on
    dummy data.

    :param pdb_df: Dataframe containing protein structure. Must contain columns
        ``["x_coord", "y_coord", "z_coord"]``.
    :type pdb_df: pd.DataFrames
    :raises: ValueError if ``pdb_df`` does not contain the required columns.
    :return: pd.Dataframe of Euclidean distance matrix.
    :rtype: pd.DataFrame
    """
    if (
        not pd.Series(["x_coord", "y_coord", "z_coord"])
        .isin(pdb_df.columns)
        .all()
    ):
        raise ValueError(
            "Dataframe must contain columns ['x_coord', 'y_coord', 'z_coord']"
        )
    eucl_dists = pdist(
        pdb_df[["x_coord", "y_coord", "z_coord"]], metric="euclidean"
    )
    eucl_dists = pd.DataFrame(squareform(eucl_dists))
    eucl_dists.index = pdb_df.index
    eucl_dists.columns = pdb_df.index

    return eucl_dists


def filter_distmat(
    pdb_df: pd.DataFrame,
    distmat: pd.DataFrame,
    exclude_edges: Iterable[str],
    inplace: bool = True,
) -> pd.DataFrame:
    """
    Filter distance matrix in place based on edge types to exclude.

    :param pdb_df: Data frame representing a PDB graph.
    :type pdb_df: pd.DataFrame
    :param distmat: Pairwise-distance matrix between all nodes
    :type pdb_df: pd.DataFrame
    :param exclude_edges: Supported values: `inter`, `intra`
        - `inter` removes inter-connections between nodes of the same chain.
        - `intra` removes intra-connections between nodes of different chains.
    :type exclude_edges: Iterable[str]
    :param inplace: False to create a deep copy.
    :type inplace: bool
    :return: Modified pairwise-distance matrix between all nodes.
    :rtype: pd.DataFrame
    """
    # Process input argument values
    supported_exclude_edges_vals = ["inter", "intra"]
    for val in exclude_edges:
        if val not in supported_exclude_edges_vals:
            raise ValueError(f"Unknown `exclude_edges` value '{val}'.")
    if not inplace:
        distmat = distmat.copy(deep=True)

    # Prepare
    chain_to_nodes = (
        pdb_df.groupby("chain_id")["node_id"].apply(list).to_dict()
    )
    node_id_to_int = dict(zip(pdb_df["node_id"], pdb_df.index))
    chain_to_nodes = {
        ch: [node_id_to_int[n] for n in nodes]
        for ch, nodes in chain_to_nodes.items()
    }

    # Construct indices of edges to exclude
    edges_to_excl = []
    if "intra" in exclude_edges:
        for nodes in chain_to_nodes.values():
            edges_to_excl.extend(list(combinations(nodes, 2)))
    if "inter" in exclude_edges:
        for nodes0, nodes1 in combinations(chain_to_nodes.values(), 2):
            edges_to_excl.extend(list(product(nodes0, nodes1)))

    # Filter distance matrix based on indices of edges to exclude
    if len(edges_to_excl):
        row_idx_to_excl, col_idx_to_excl = zip(*edges_to_excl)
        distmat.iloc[row_idx_to_excl, col_idx_to_excl] = INFINITE_DIST
        distmat.iloc[col_idx_to_excl, row_idx_to_excl] = INFINITE_DIST

    return distmat


def add_edge(G, n1, n2, kind_name):
    if G.has_edge(n1, n2):
        G.edges[n1, n2]["kind"].add(kind_name)
    else:
        G.add_edge(n1, n2, kind={kind_name})


def add_distance_to_edges(G: nx.Graph) -> nx.Graph:
    """Adds Euclidean distance between nodes in an edge as an edge attribute.

    :param G: Graph to add distances to.
    :type G: nx.Graph
    :return: Graph with added distances.
    :rtype: nx.Graph
    """
    if "atomic_dist_mat" in G.graph.keys():
        dist_mat = G.graph["atomic_dist_mat"]
    elif "dist_mat" in G.graph.keys():
        dist_mat = G.graph["dist_mat"]
    else:
        dist_mat = compute_distmat(G.graph["pdb_df"])
        G.graph["dist_mat"] = dist_mat

    mat = np.where(nx.to_numpy_array(G), dist_mat, 0)
    node_map = {n: i for i, n in enumerate(G.nodes)}
    for u, v, d in G.edges(data=True):
        d["distance"] = mat[node_map[u], node_map[v]]
    return G


def add_sequence_distance_edges(
    G: nx.Graph, d: int, name: str = "sequence_edge"
) -> nx.Graph:
    """
    Adds edges based on sequence distance to residues in each chain.

    Eg. if ``d=6`` then we join: nodes ``(1,7), (2,8), (3,9)..``
    based on their sequence number.

    :param G: Networkx protein graph.
    :type G: nx.Graph
    :param d: Sequence separation to add edges on.
    :param name: Name of the edge type. Defaults to ``"sequence_edge"``.
    :type name: str
    :return G: Networkx protein graph with added peptide bonds.
    :rtype: nx.Graph
    """
    # Iterate over every chain
    for chain_id in G.graph["chain_ids"]:
        # Find chain residues
        chain_residues = [
            (n, v) for n, v in G.nodes(data=True) if v["chain_id"] == chain_id
        ]

        # Subset to only N and C atoms in the case of full-atom
        # peptide bond addition
        try:
            if (
                G.graph["config"].granularity == "atom"
                and name == "peptide_bond"
            ):
                chain_residues = [
                    (n, v)
                    for n, v in chain_residues
                    if v["atom_type"] in {"N", "C"}
                ]
        # If we don't don't find a config, assume it's a residue graph
        # This is brittle
        except KeyError:
            continue

        # Iterate over every residue in chain
        for i, residue in enumerate(chain_residues):
            try:
                # Checks not at chain terminus - is this versatile enough?
                if i == len(chain_residues) - d:
                    continue
                # Asserts residues are on the same chain
                cond_1 = (
                    residue[1]["chain_id"]
                    == chain_residues[i + d][1]["chain_id"]
                )
                # Asserts residue numbers are adjacent
                cond_2 = (
                    abs(
                        residue[1]["residue_number"]
                        - chain_residues[i + d][1]["residue_number"]
                    )
                    == d
                )

                # If this checks out, we add a peptide bond
                if (cond_1) and (cond_2):
                    # Adds "peptide bond" between current residue and the next
                    if G.has_edge(i, i + d):
                        G.edges[i, i + d]["kind"].add(name)
                    else:
                        G.add_edge(
                            residue[0],
                            chain_residues[i + d][0],
                            kind={name},
                        )
            except IndexError:
                continue
    return G


def add_peptide_bonds(G: nx.Graph) -> nx.Graph:
    """
    Adds peptide backbone as edges to residues in each chain.

    :param G: Networkx protein graph.
    :type G: nx.Graph
    :return G: Networkx protein graph with added peptide bonds.
    :rtype: nx.Graph
    """
    return add_sequence_distance_edges(G, d=1, name="peptide_bond")


def add_hydrophobic_interactions(
    G: nx.Graph, rgroup_df: Optional[pd.DataFrame] = None
):
    """
    Find all hydrophobic interactions.

    Performs searches between the following residues:
    ``[ALA, VAL, LEU, ILE, MET, PHE, TRP, PRO, TYR]``
    (:const:`~graphein.protein.resi_atoms.HYDROPHOBIC_RESIS`).

    Criteria: R-group residues are within 5A distance.

    :param G: nx.Graph to add hydrophobic interactions to.
    :type G: nx.Graph
    :param rgroup_df: Optional dataframe of R-group atoms.
    :type rgroup_df: pd.DataFrame, optional
    """
    if rgroup_df is None:
        rgroup_df = G.graph["rgroup_df"]
    hydrophobics_df = filter_dataframe(
        rgroup_df, "residue_name", HYDROPHOBIC_RESIS, True
    )
    hydrophobics_df = filter_dataframe(
        hydrophobics_df, "node_id", list(G.nodes()), True
    )
    if hydrophobics_df.shape[0] > 0:
        distmat = compute_distmat(hydrophobics_df)
        interacting_atoms = get_interacting_atoms(5, distmat)
        add_interacting_resis(
            G, interacting_atoms, hydrophobics_df, ["hydrophobic"]
        )


def add_disulfide_interactions(
    G: nx.Graph, rgroup_df: Optional[pd.DataFrame] = None
):
    """
    Find all disulfide interactions between CYS residues
    (:const:`~graphein.protein.resi_atoms.DISULFIDE_RESIS`,
    :const:`~graphein.protein.resi_atoms.DISULFIDE_ATOMS`).

    Criteria: sulfur atom pairs are within 2.2A of each other.

    :param G: networkx protein graph
    :type G: nx.Graph
    :param rgroup_df: pd.DataFrame containing rgroup data, defaults to ``None``,
        which retrieves the df from the provided nx graph.
    :type rgroup_df: pd.DataFrame, optional
    """
    # Check for existence of at least two Cysteine residues
    residues = [d["residue_name"] for _, d in G.nodes(data=True)]
    if residues.count("CYS") < 2:
        log.debug(
            f"{residues.count('CYS')} CYS residues found. Cannot add disulfide \
                interactions with fewer than two CYS residues."
        )
        return

    if rgroup_df is None:
        rgroup_df = G.graph["rgroup_df"]
    disulfide_df = filter_dataframe(
        rgroup_df, "residue_name", DISULFIDE_RESIS, True
    )
    disulfide_df = filter_dataframe(
        disulfide_df, "atom_name", DISULFIDE_ATOMS, True
    )
    # Ensure only residues in the graph are kept
    disulfide_df = filter_dataframe(
        disulfide_df, "node_id", list(G.nodes), True
    )
    if disulfide_df.shape[0] > 0:
        distmat = compute_distmat(disulfide_df)
        interacting_atoms = get_interacting_atoms(2.2, distmat)
        add_interacting_resis(
            G, interacting_atoms, disulfide_df, ["disulfide"]
        )


def add_hydrogen_bond_interactions(
    G: nx.Graph, rgroup_df: Optional[pd.DataFrame] = None
):
    """Add all hydrogen-bond interactions."""
    # For these atoms, find those that are within 3.5A of one another.
    if rgroup_df is None:
        rgroup_df = G.graph["rgroup_df"]
    rgroup_df = filter_dataframe(rgroup_df, "node_id", list(G.nodes()), True)
    HBOND_ATOMS = [
        "ND",  # histidine and asparagine
        "NE",  # glutamate, tryptophan, arginine, histidine
        "NH",  # arginine
        "NZ",  # lysine
        "OD1",
        "OD2",
        "OE",
        "OG",
        "OH",
        "SD",  # cysteine
        "SG",  # methionine
        "N",
        "O",
    ]
    hbond_df = filter_dataframe(rgroup_df, "atom_name", HBOND_ATOMS, True)
    if len(hbond_df.index) > 0:
        distmat = compute_distmat(hbond_df)
        interacting_atoms = get_interacting_atoms(3.5, distmat)
        add_interacting_resis(G, interacting_atoms, hbond_df, ["hbond"])

    # For these atoms, find those that are within 4.0A of one another.
    HBOND_ATOMS_SULPHUR = ["SD", "SG"]
    hbond_df = filter_dataframe(
        rgroup_df, "atom_name", HBOND_ATOMS_SULPHUR, True
    )
    if len(hbond_df.index) > 0:
        distmat = compute_distmat(hbond_df)
        interacting_atoms = get_interacting_atoms(4.0, distmat)
        add_interacting_resis(G, interacting_atoms, hbond_df, ["hbond"])


def add_ionic_interactions(
    G: nx.Graph, rgroup_df: Optional[pd.DataFrame] = None
):
    """
    Find all ionic interactions.

    Criteria: ``[ARG, LYS, HIS, ASP, and GLU]``
    (:const:`~graphein.protein.resi_atoms.IONIC_RESIS`) residues are within 6A.

    We also check for opposing charges
    (:const:`~graphein.protein.resi_atoms.POS_AA`,
    :const:`~graphein.protein.resi_atoms.NEG_AA`).

    :param G: nx.Graph to add ionic interactions to.
    :type G: nx.Graph
    :param rgroup_df: Optional dataframe of R-group atoms. Default is ``None``.
    :type rgroup_df: Optional[pd.DataFrame]
    """
    if rgroup_df is None:
        rgroup_df = G.graph["rgroup_df"]
    ionic_df = filter_dataframe(rgroup_df, "residue_name", IONIC_RESIS, True)
    ionic_df = filter_dataframe(rgroup_df, "node_id", list(G.nodes()), True)
    if ionic_df.shape[0] > 0:
        distmat = compute_distmat(ionic_df)
        interacting_atoms = get_interacting_atoms(6, distmat)
        add_interacting_resis(G, interacting_atoms, ionic_df, ["ionic"])
        # Check that the interacting residues are of opposite charges
        for r1, r2 in get_edges_by_bond_type(G, "ionic"):
            condition1 = (
                G.nodes[r1]["residue_name"] in POS_AA
                and G.nodes[r2]["residue_name"] in NEG_AA
            )
            condition2 = (
                G.nodes[r2]["residue_name"] in POS_AA
                and G.nodes[r1]["residue_name"] in NEG_AA
            )
            is_ionic = condition1 or condition2
            if not is_ionic:
                G.edges[r1, r2]["kind"].remove("ionic")
                if len(G.edges[r1, r2]["kind"]) == 0:
                    G.remove_edge(r1, r2)


def add_aromatic_interactions(
    G: nx.Graph, pdb_df: Optional[pd.DataFrame] = None
):
    """
    Find all aromatic-aromatic interaction.

    Criteria: phenyl ring centroids separated between 4.5A to 7A.
    Phenyl rings are present on ``PHE, TRP, HIS, TYR``
    (:const:`~graphein.protein.resi_atoms.AROMATIC_RESIS`).
    Phenyl ring atoms on these amino acids are defined by the following
    atoms:
    - PHE: CG, CD, CE, CZ
    - TRP: CD, CE, CH, CZ
    - HIS: CG, CD, ND, NE, CE
    - TYR: CG, CD, CE, CZ
    Centroids of these atoms are taken by taking:
        (mean x), (mean y), (mean z)
    for each of the ring atoms.
    Notes for future self/developers:
    - Because of the requirement to pre-compute ring centroids, we do not
        use the functions written above (filter_dataframe, compute_distmat,
        get_interacting_atoms), as they do not return centroid atom
        euclidean coordinates.
    """
    if pdb_df is None:
        pdb_df = G.graph["raw_pdb_df"]
    dfs = []
    for resi in AROMATIC_RESIS:
        resi_rings_df = get_ring_atoms(pdb_df, resi)
        resi_rings_df = filter_dataframe(
            resi_rings_df, "node_id", list(G.nodes()), True
        )
        resi_centroid_df = get_ring_centroids(resi_rings_df)
        dfs.append(resi_centroid_df)
    aromatic_df = (
        pd.concat(dfs).sort_values(by="node_id").reset_index(drop=True)
    )
    if aromatic_df.shape[0] > 0:
        distmat = compute_distmat(aromatic_df)
        distmat.set_index(aromatic_df["node_id"], inplace=True)
        distmat.columns = aromatic_df["node_id"]
        distmat = distmat[(distmat >= 4.5) & (distmat <= 7)].fillna(0)
        indices = np.where(distmat > 0)

        interacting_resis = [
            (distmat.index[r], distmat.index[c])
            for r, c in zip(indices[0], indices[1])
        ]
        log.info(
            f"Found: {len(interacting_resis)} aromatic-aromatic interactions"
        )
        for n1, n2 in interacting_resis:
            assert G.nodes[n1]["residue_name"] in AROMATIC_RESIS
            assert G.nodes[n2]["residue_name"] in AROMATIC_RESIS
            add_edge(G, n1, n2, "aromatic")


def add_aromatic_sulphur_interactions(
    G: nx.Graph, rgroup_df: Optional[pd.DataFrame] = None
):
    """Find all aromatic-sulphur interactions.

    Criteria: Sulphur containing residue () within 5.3 Angstroms of an aromatic
    residue ().

    :param G: The graph to add the aromatic-sulphur interactions to.
    :type G: nx.Graph
    :param rgroup_df: The rgroup dataframe. If ``None`` (default), the graph's
        rgroup dataframe is used.
    :type rgroup_df: Optional[pd.DataFrame].
    """
    if rgroup_df is None:
        rgroup_df = G.graph["rgroup_df"]
    RESIDUES = SULPHUR_RESIS + PI_RESIS

    aromatic_sulphur_df = filter_dataframe(
        rgroup_df, "residue_name", RESIDUES, True
    )
    aromatic_sulphur_df = filter_dataframe(
        aromatic_sulphur_df, "node_id", list(G.nodes()), True
    )

    if aromatic_sulphur_df.shape[0] > 0:
        distmat = compute_distmat(aromatic_sulphur_df)
        interacting_atoms = get_interacting_atoms(5.3, distmat)
        interacting_atoms = list(
            zip(interacting_atoms[0], interacting_atoms[1])
        )

        for a1, a2 in interacting_atoms:
            resi1 = aromatic_sulphur_df.loc[a1, "node_id"]
            resi2 = aromatic_sulphur_df.loc[a2, "node_id"]

            condition1 = resi1 in SULPHUR_RESIS and resi2 in PI_RESIS
            condition2 = resi1 in PI_RESIS and resi2 in SULPHUR_RESIS

            if (condition1 or condition2) and resi1 != resi2:
                add_edge(G, resi1, resi2, "aromatic_sulphur")


def add_cation_pi_interactions(
    G: nx.Graph, rgroup_df: Optional[pd.DataFrame] = None
):
    """Add cation-pi interactions.

    Criteria:
        # Todo

    :param G: Graph to add cation-pi interactions to.
    :type G: nx.Graph
    :param rgroup_df: Dataframe containing rgroup information. Defaults to
        ``None``.
    :type rgroup_df: Optional[pd.DataFrame].
    """
    if rgroup_df is None:
        rgroup_df = G.graph["rgroup_df"]
    cation_pi_df = filter_dataframe(
        rgroup_df, "residue_name", CATION_PI_RESIS, True
    )
    cation_pi_df = filter_dataframe(
        cation_pi_df, "node_id", list(G.nodes()), True
    )

    if cation_pi_df.shape[0] > 0:
        distmat = compute_distmat(cation_pi_df)
        interacting_atoms = get_interacting_atoms(6, distmat)
        interacting_atoms = list(
            zip(interacting_atoms[0], interacting_atoms[1])
        )

        for a1, a2 in interacting_atoms:
            resi1 = cation_pi_df.loc[a1, "node_id"]
            resi2 = cation_pi_df.loc[a2, "node_id"]

            condition1 = resi1 in CATION_RESIS and resi2 in PI_RESIS
            condition2 = resi1 in PI_RESIS and resi2 in CATION_RESIS

            if (condition1 or condition2) and resi1 != resi2:
                add_edge(G, resi1, resi2, "cation_pi")


def add_vdw_interactions(
    g: nx.Graph,
    threshold: float = 0.5,
    remove_intraresidue: bool = False,
    name: str = "vdw",
):
    """Criterion: Any non-H atoms within the sum of
    their VdW Radii (:const:`~graphein.protein.resi_atoms.VDW_RADII`) +
    threshold (default: ``0.5``) Angstroms of each other.

    :param g: Graph to add van der Waals interactions to.
    :type g: nx.Graph
    :param threshold: Threshold distance for van der Waals interactions.
        Default: ``0.5`` Angstroms.
    :type threshold: float
    :param remove_intraresidue: Whether to remove intra-residue interactions.
    :type remove_intraresidue: bool
    """
    df = g.graph["raw_pdb_df"]
    df = filter_dataframe(df, "atom_name", ["H"], boolean=False)
    df = filter_dataframe(df, "node_id", list(g.nodes()), boolean=True)
    dist_mat = compute_distmat(df)

    radii = df["element_symbol"].map(VDW_RADII).values
    radii = np.expand_dims(radii, axis=1)
    radii = radii + radii.T

    dist_mat = dist_mat - radii
    interacting_atoms = get_interacting_atoms(threshold, dist_mat)
    add_interacting_resis(g, interacting_atoms, df, [name])

    if remove_intraresidue:
        for u, v in get_edges_by_bond_type(g, name):
            u_id = "".join(u.split(":")[:-1])
            v_id = "".join(v.split(":")[:-1])
            if u_id == v_id:
                g.edges[u, v]["kind"].remove(name)
                if len(g.edges[u, v]["kind"]) == 0:
                    g.remove_edge(u, v)


def add_vdw_clashes(
    g: nx.Graph, threshold: float = 0.0, remove_intraresidue: bool = False
):
    """Adds van der Waals clashes to graph.

    These are atoms that are within the sum of their VdW Radii
    (:const:`~graphein.protein.resi_atoms.VDW_RADII`).

    :param g: Graph to add van der Waals clashes to.
    :type g: nx.Graph
    :param threshold: Threshold, defaults to ``0.0``.
    :type threshold: float
    :param remove_intraresidue: Whether to remove clashes within a residue,
        defaults to ``False``.
    :type remove_intraresidue: bool
    """
    add_vdw_interactions(
        g,
        threshold=threshold,
        remove_intraresidue=remove_intraresidue,
        name="vdw_clash",
    )


def add_pi_stacking_interactions(
    G: nx.Graph,
    pdb_df: Optional[pd.DataFrame] = None,
    centroid_distance: float = 7.0,
):
    """Adds Pi-stacking interactions to graph.

    Criteria:
        - aromatic ring centroids within 7.0 (default) Angstroms. (|A1A2| < 7.0)
        - Angle between ring normal vectors < 30° (∠(n1, n2) < 30°)
        - Angle between ring normal vectors and centroid vector < 45°
            (∠(n1, A1A2) < 45°), (∠(n2, A1A2) < 45°)

    :param G: _description_
    :type G: nx.Graph
    :param pdb_df: _description_, defaults to None
    :type pdb_df: Optional[pd.DataFrame], optional
    """
    if pdb_df is None:
        pdb_df = G.graph["raw_pdb_df"]
    dfs = []
    # Compute centroids and normal for each ring
    for resi in PI_RESIS:
        resi_rings_df = get_ring_atoms(pdb_df, resi)
        resi_rings_df = filter_dataframe(
            resi_rings_df, "node_id", list(G.nodes()), True
        )
        resi_centroid_df = get_ring_centroids(resi_rings_df)
        resi_normals = get_ring_normals(resi_rings_df)
        resi_centroid_df = pd.merge(
            resi_centroid_df, resi_normals, on="node_id"
        )
        dfs.append(resi_centroid_df)
    aromatic_df = (
        pd.concat(dfs).sort_values(by="node_id").reset_index(drop=True)
    )

    if aromatic_df.shape[0] > 0:
        distmat = compute_distmat(aromatic_df)
        distmat.set_index(aromatic_df["node_id"], inplace=True)
        distmat.columns = aromatic_df["node_id"]
        distmat = distmat[distmat <= centroid_distance].fillna(0)
        indices = np.where(distmat > 0)
        interacting_resis = [
            (distmat.index[r], distmat.index[c])
            for r, c in zip(indices[0], indices[1])
        ]
        # log.info(f"Found: {len(interacting_resis)} aromatic-aromatic interactions")
        for n1, n2 in interacting_resis:
            assert G.nodes[n1]["residue_name"] in PI_RESIS
            assert G.nodes[n2]["residue_name"] in PI_RESIS
            n1_centroid = aromatic_df.loc[aromatic_df["node_id"] == n1][
                ["x_coord", "y_coord", "z_coord"]
            ].values[0]
            n2_centroid = aromatic_df.loc[aromatic_df["node_id"] == n2][
                ["x_coord", "y_coord", "z_coord"]
            ].values[0]

            n1_normal = aromatic_df.loc[aromatic_df["node_id"] == n1][
                0
            ].values[0]
            n2_normal = aromatic_df.loc[aromatic_df["node_id"] == n2][
                0
            ].values[0]

            centroid_vector = n2_centroid - n1_centroid

            norm_angle = compute_angle(n1_normal, n2_normal)
            n1_centroid_angle = compute_angle(n1_normal, centroid_vector)
            n2_centroid_angle = compute_angle(n2_normal, centroid_vector)

            if (
                norm_angle >= 30
                or n1_centroid_angle >= 45
                or n2_centroid_angle >= 45
            ):
                continue
            add_edge(G, n1, n2, "pi_stacking")


def add_t_stacking(G: nx.Graph, pdb_df: Optional[pd.DataFrame] = None):
    if pdb_df is None:
        pdb_df = G.graph["raw_pdb_df"]
    dfs = []
    # Compute centroids and normal for each ring
    for resi in PI_RESIS:
        resi_rings_df = get_ring_atoms(pdb_df, resi)
        resi_rings_df = filter_dataframe(
            resi_rings_df, "node_id", list(G.nodes()), True
        )
        resi_centroid_df = get_ring_centroids(resi_rings_df)
        resi_normals = get_ring_normals(resi_rings_df)
        resi_centroid_df = pd.merge(
            resi_centroid_df, resi_normals, on="node_id"
        )
        dfs.append(resi_centroid_df)
    aromatic_df = (
        pd.concat(dfs).sort_values(by="node_id").reset_index(drop=True)
    )

    if aromatic_df.shape[0] > 0:
        distmat = compute_distmat(aromatic_df)
        distmat.set_index(aromatic_df["node_id"], inplace=True)
        distmat.columns = aromatic_df["node_id"]
        distmat = distmat[distmat <= 7].fillna(0)
        indices = np.where(distmat > 0)
        interacting_resis = [
            (distmat.index[r], distmat.index[c])
            for r, c in zip(indices[0], indices[1])
        ]
        # log.info(f"Found: {len(interacting_resis)} aromatic-aromatic interactions")
        for n1, n2 in interacting_resis:
            assert G.nodes[n1]["residue_name"] in PI_RESIS
            assert G.nodes[n2]["residue_name"] in PI_RESIS
            n1_centroid = aromatic_df.loc[aromatic_df["node_id"] == n1][
                ["x_coord", "y_coord", "z_coord"]
            ].values[0]
            n2_centroid = aromatic_df.loc[aromatic_df["node_id"] == n2][
                ["x_coord", "y_coord", "z_coord"]
            ].values[0]

            n1_normal = aromatic_df.loc[aromatic_df["node_id"] == n1][
                0
            ].values[0]
            n2_normal = aromatic_df.loc[aromatic_df["node_id"] == n2][
                0
            ].values[0]

            centroid_vector = n2_centroid - n1_centroid

            norm_angle = compute_angle(n1_normal, n2_normal)
            n1_centroid_angle = compute_angle(n1_normal, centroid_vector)
            n2_centroid_angle = compute_angle(n2_normal, centroid_vector)

            if (
                norm_angle >= 90
                or norm_angle <= 60
                or n1_centroid_angle >= 45
                or n2_centroid_angle >= 45
            ):
                continue
            add_edge(G, n1, n2, "t_stacking")


def add_backbone_carbonyl_carbonyl_interactions(
    G: nx.Graph, threshold: float = 3.2
):
    """Adds backbone-carbonyl-carbonyl interactions.

    Default is to consider C═O···C═O interactions below 3.2 Angstroms
    (sum of O+C vdw radii).

    Source:
    > Rahim, A., Saha, P., Jha, K.K. et al. Reciprocal carbonyl–carbonyl
    > interactions in small molecules and proteins. Nat Commun 8, 78 (2017).
    > https://doi.org/10.1038/s41467-017-00081-x

    :param G: Protein graph to add edges to.
    :type G: nx.Graph
    :param threshold: Threshold below which to consider an interaction,
        defaults to 3.2 Angstroms.
    :type threshold: float, optional
    """
    df = G.graph["raw_pdb_df"]
    df = filter_dataframe(df, "node_id", list(G.nodes()), boolean=True)
    df = filter_dataframe(df, "atom_name", ["C", "O"], boolean=True)
    distmat = compute_distmat(df)
    interacting_atoms = get_interacting_atoms(threshold, distmat)

    # Filter out O-O and C-C edges
    atom_1 = df.iloc[interacting_atoms[0]]["atom_name"]
    atom_2 = df.iloc[interacting_atoms[1]]["atom_name"]
    diff_atoms = atom_1.values != atom_2.values
    interacting_atoms = (
        interacting_atoms[0][diff_atoms],
        interacting_atoms[1][diff_atoms],
    )

    # Filter out O-C edges on the same residue
    atom_1 = df.iloc[interacting_atoms[0]]["residue_id"]
    atom_2 = df.iloc[interacting_atoms[1]]["residue_id"]
    diff_atoms = atom_1.values != atom_2.values
    interacting_atoms = (
        interacting_atoms[0][diff_atoms],
        interacting_atoms[1][diff_atoms],
    )

    add_interacting_resis(G, interacting_atoms, df, ["bb_carbonyl_carbonyl"])


def add_salt_bridges(
    G: nx.Graph,
    rgroup_df: Optional[pd.DataFrame] = None,
    threshold: float = 4.0,
):
    """Compute salt bridge interactions.

    Criterion: Anion-Cation residue atom pairs within threshold (``4.0``)
    Angstroms of each other.

    Anions: ASP/OD1+OD2, GLU/OE1+OE2
    Cations: LYS/NZ, ARG/NH1+NH2

    :param G: Graph to add salt bridge interactions to.
    :type G: nx.Graph
    :param rgroup_df: R group dataframe, defaults to ``None``.
    :type rgroup_df: Optional[pd.DataFrame]
    :param threshold: Distance threshold, defaults to ``4.0`` Angstroms.
    :type threshold: float, optional
    """
    if rgroup_df is None:
        rgroup_df = G.graph["rgroup_df"]
    salt_bridge_df = filter_dataframe(
        rgroup_df, "residue_name", SALT_BRIDGE_RESIDUES, boolean=True
    )
    salt_bridge_df = filter_dataframe(
        salt_bridge_df, "atom_name", SALT_BRIDGE_ATOMS, boolean=True
    )
    if salt_bridge_df.shape[0] > 0:
        distmat = compute_distmat(salt_bridge_df)
        interacting_atoms = get_interacting_atoms(threshold, distmat)
        add_interacting_resis(
            G, interacting_atoms, salt_bridge_df, ["salt_bridge"]
        )

        for r1, r2 in get_edges_by_bond_type(G, "salt_bridge"):
            condition1 = (
                G.nodes[r1]["residue_name"] in SALT_BRIDGE_ANIONS
                and G.nodes[r2]["residue_name"] in SALT_BRIDGE_CATIONS
            )
            condition2 = (
                G.nodes[r2]["residue_name"] in SALT_BRIDGE_ANIONS
                and G.nodes[r1]["residue_name"] in SALT_BRIDGE_CATIONS
            )
            is_ionic = condition1 or condition2
            if not is_ionic:
                G.edges[r1, r2]["kind"].remove("salt_bridge")
                if len(G.edges[r1, r2]["kind"]) == 0:
                    G.remove_edge(r1, r2)


def get_interacting_atoms(
    angstroms: float, distmat: pd.DataFrame
) -> np.ndarray:
    """Find the atoms that are within a particular radius of one another.

    :param angstroms: The radius in angstroms.
    :type angstroms: float
    :param distmat: The distance matrix.
    :type distmat: pd.DataFrame
    """
    return np.where(distmat <= angstroms)


def add_delaunay_triangulation(
    G: nx.Graph, allowable_nodes: Optional[List[str]] = None
):
    """
    Compute the Delaunay triangulation of the protein structure.

    This has been used in prior work. References:

        Harrison, R. W., Yu, X. & Weber, I. T. Using triangulation to include
        target structure improves drug resistance prediction accuracy. in 1–1
        (IEEE, 2013). doi:10.1109/ICCABS.2013.6629236

        Yu, X., Weber, I. T. & Harrison, R. W. Prediction of HIV drug
        resistance from genotype with encoded three-dimensional protein
        structure. BMC Genomics 15 Suppl 5, S1 (2014).

    Notes:
    1. We do not use the add_interacting_resis function, because this
        interaction is computed on the ``CA`` atoms. Therefore, there is code
        duplication. For now, I have chosen to leave this code duplication
        in.

    :param G: The networkx graph to add the triangulation to.
    :type G: nx.Graph
    :param allowable_nodes: The nodes to include in the triangulation.
        If ``None`` (default), no filtering is done. This parameter is used to
        filter out nodes that are not desired in the triangulation. Eg if you
        wanted to construct a delaunay triangulation of the CA atoms of an
        atomic graph.
    :type allowable_nodes: List[str], optional
    """
    if allowable_nodes is None:
        coords = np.array([d["coords"] for _, d in G.nodes(data=True)])
        node_map: Dict[int, str] = dict(enumerate(G.nodes()))
    else:
        coords = np.array(
            [
                d["coords"]
                for _, d in G.nodes(data=True)
                if d["atom_type"] in allowable_nodes
            ]
        )
        node_map: Dict[int, str] = {
            i: n
            for i, (n, d) in enumerate(G.nodes(data=True))
            if d["atom_type"] in allowable_nodes
        }
        node_map: Dict[int, str] = dict(enumerate(node_map.values()))

    tri = Delaunay(coords)  # this is the triangulation
    log.debug(
        f"Detected {len(tri.simplices)} simplices in the Delaunay Triangulation."
    )
    for simplex in tri.simplices:
        nodes = [node_map[s] for s in simplex]
        for n1, n2 in combinations(nodes, 2):
            if n1 not in G.nodes or n2 not in G.nodes:
                continue
            add_edge(G, n1, n2, "delaunay")


def add_distance_threshold(
    G: nx.Graph, long_interaction_threshold: int, threshold: float = 5.0
):
    """
    Adds edges to any nodes within a given distance of each other.
    Long interaction threshold is used to specify minimum separation in sequence
    to add an edge between networkx nodes within the distance threshold

    :param G: Protein Structure graph to add distance edges to
    :type G: nx.Graph
    :param long_interaction_threshold: minimum distance in sequence for two
        nodes to be connected
    :type long_interaction_threshold: int
    :param threshold: Distance in angstroms, below which two nodes are connected
    :type threshold: float
    :return: Graph with distance-based edges added
    """
    pdb_df = filter_dataframe(
        G.graph["pdb_df"], "node_id", list(G.nodes()), True
    )
    dist_mat = compute_distmat(pdb_df)
    interacting_nodes = get_interacting_atoms(threshold, distmat=dist_mat)
    interacting_nodes = list(zip(interacting_nodes[0], interacting_nodes[1]))

    log.info(f"Found: {len(interacting_nodes)} distance edges")
    count = 0
    for a1, a2 in interacting_nodes:
        n1 = G.graph["pdb_df"].loc[a1, "node_id"]
        n2 = G.graph["pdb_df"].loc[a2, "node_id"]
        n1_chain = G.graph["pdb_df"].loc[a1, "chain_id"]
        n2_chain = G.graph["pdb_df"].loc[a2, "chain_id"]
        n1_position = G.graph["pdb_df"].loc[a1, "residue_number"]
        n2_position = G.graph["pdb_df"].loc[a2, "residue_number"]

        condition_1 = n1_chain == n2_chain
        condition_2 = (
            abs(n1_position - n2_position) < long_interaction_threshold
        )

        if not (condition_1 and condition_2):
            count += 1
            add_edge(G, n1, n2, "distance_threshold")

    log.info(
        f"Added {count} distance edges. ({len(list(interacting_nodes)) - count}\
            removed by LIN)"
    )


def add_distance_window(
    G: nx.Graph, min: float, max: float, long_interaction_threshold: int = -1
):
    """
    Adds edges to any nodes within a given window of distances of each other.
    Long interaction threshold is used
    to specify minimum separation in sequence to add an edge between networkx
    nodes within the distance threshold

    :param G: Protein Structure graph to add distance edges to
    :type G: nx.Graph
    :param min: Minimum distance in angstroms required for an edge.
    :type min: float
    :param max: Maximum distance in angstroms allowed for an edge.
    :param long_interaction_threshold: minimum distance in sequence for two
        nodes to be connected
    :type long_interaction_threshold: int
    :return: Graph with distance-based edges added
    """
    pdb_df = filter_dataframe(
        G.graph["pdb_df"], "node_id", list(G.nodes()), True
    )
    dist_mat = compute_distmat(pdb_df)
    # Nodes less than the minimum distance
    less_than_min = get_interacting_atoms(min, distmat=dist_mat)
    less_than_min = list(zip(less_than_min[0], less_than_min[1]))

    interacting_nodes = get_interacting_atoms(max, distmat=dist_mat)
    interacting_nodes = list(zip(interacting_nodes[0], interacting_nodes[1]))
    interacting_nodes = [
        i for i in interacting_nodes if i not in less_than_min
    ]

    log.info(f"Found: {len(interacting_nodes)} distance edges")
    count = 0
    for a1, a2 in interacting_nodes:
        n1 = G.graph["pdb_df"].loc[a1, "node_id"]
        n2 = G.graph["pdb_df"].loc[a2, "node_id"]
        n1_chain = G.graph["pdb_df"].loc[a1, "chain_id"]
        n2_chain = G.graph["pdb_df"].loc[a2, "chain_id"]
        n1_position = G.graph["pdb_df"].loc[a1, "residue_number"]
        n2_position = G.graph["pdb_df"].loc[a2, "residue_number"]

        condition_1 = n1_chain == n2_chain
        condition_2 = (
            abs(n1_position - n2_position) < long_interaction_threshold
        )

        if not (condition_1 and condition_2):
            count += 1
            add_edge(G, n1, n2, f"distance_window_{min}_{max}")
    log.info(
        f"Added {count} distance edges. ({len(list(interacting_nodes)) - count}\
            removed by LIN)"
    )


def add_fully_connected_edges(G: nx.Graph):
    """
    Adds fully connected edges to nodes.

    :param G: Protein structure graph to add fully connected edges to.
    :type G: nx.Graph
    """
    for n1, n2 in itertools.product(G.nodes(), G.nodes()):
        add_edge(G, n1, n2, f"fully_connected")


# TODO Support for directed edges
def add_k_nn_edges(
    G: nx.Graph,
    long_interaction_threshold: int = 0,
    k: int = 5,
    exclude_edges: Iterable[str] = (),
    exclude_self_loops: bool = True,
    kind_name: str = "knn",
):
    """
    Adds edges to nodes based on K nearest neighbours. Long interaction
    threshold is used to specify minimum separation in sequence to add an edge
    between networkx nodes within the distance threshold

    :param G: Protein Structure graph to add distance edges to
    :type G: nx.Graph
    :param long_interaction_threshold: minimum distance in sequence for two
        nodes to be connected
    :type long_interaction_threshold: int
    :param k: Number of neighbors for each sample.
    :type k: int
    :param exclude_edges: Types of edges to exclude. Supported values are
        `inter` and `intra`.
        - `inter` removes inter-connections between nodes of the same chain.
        - `intra` removes intra-connections between nodes of different chains.
    :type exclude_edges: Iterable[str].
    :param exclude_self_loops: Whether or not to mark each sample as the first
        nearest neighbor to itself.
    :type exclude_self_loops: Union[bool, str]
    :param kind_name: Name for kind of edges in networkx graph.
    :type kind_name: str
    :return: Graph with knn-based edges added
    :rtype: nx.Graph
    """
    # Prepare dataframe
    pdb_df = filter_dataframe(
        G.graph["pdb_df"], "node_id", list(G.nodes()), True
    )
    if (
        pdb_df["x_coord"].isna().sum()
        or pdb_df["y_coord"].isna().sum()
        or pdb_df["z_coord"].isna().sum()
    ):
        raise ValueError("Coordinates contain a NaN value.")
    pdb_df = pdb_df.reset_index(drop=True)

    # Construct distance matrix
    dist_mat = compute_distmat(pdb_df)

    # Filter edges
    dist_mat = filter_distmat(pdb_df, dist_mat, exclude_edges)

    # Add self-loops if specified
    if not exclude_self_loops:
        k -= 1
        for n1, n2 in zip(G.nodes(), G.nodes()):
            add_edge(G, n1, n2, kind_name)

    # Reduce k if number of nodes is less (to avoid sklearn error)
    # Note: - 1 because self-loops are not included
    if G.number_of_nodes() - 1 < k:
        k = G.number_of_nodes() - 1

    if k == 0:
        return

    # Run k-NN search
    neigh = NearestNeighbors(n_neighbors=k, metric="precomputed")
    neigh.fit(dist_mat)
    nn = neigh.kneighbors_graph()

    # Create iterable of node indices
    outgoing = np.repeat(np.array(range(len(pdb_df))), k)
    incoming = nn.indices
    interacting_nodes = list(zip(outgoing, incoming))
    log.info(f"Found: {len(interacting_nodes)} KNN edges")
    for a1, a2 in interacting_nodes:
        if dist_mat.loc[a1, a2] == INFINITE_DIST:
            continue

        # Get nodes IDs from indices
        n1 = pdb_df.loc[a1, "node_id"]
        n2 = pdb_df.loc[a2, "node_id"]

        # Get chains
        n1_chain = pdb_df.loc[a1, "chain_id"]
        n2_chain = pdb_df.loc[a2, "chain_id"]

        # Get sequence position
        n1_position = pdb_df.loc[a1, "residue_number"]
        n2_position = pdb_df.loc[a2, "residue_number"]

        # Check residues are not on same chain
        condition_1 = n1_chain != n2_chain
        # Check residues are separated by long_interaction_threshold
        condition_2 = (
            abs(n1_position - n2_position) > long_interaction_threshold
        )

        # If not on same chain add edge or
        # If on same chain and separation is sufficient add edge
        if condition_1 or condition_2:
            add_edge(G, n1, n2, kind_name)


def get_ring_atoms(dataframe: pd.DataFrame, aa: str) -> pd.DataFrame:
    """
    Return ring atoms from a dataframe.

    A helper function for add_aromatic_interactions.

    Gets the ring atoms from the particular aromatic amino acid.

    Parameters:
    ===========
    - dataframe: the dataframe containing the atom records.
    - aa: the amino acid of interest, passed in as 3-letter string.

    Returns:
    ========
    - dataframe: a filtered dataframe containing just those atoms from the
                    particular amino acid selected. e.g. equivalent to
                    selecting just the ring atoms from a particular amino
                    acid.
    """
    ring_atom_df = filter_dataframe(dataframe, "residue_name", [aa], True)

    ring_atom_df = filter_dataframe(
        ring_atom_df, "atom_name", AA_RING_ATOMS[aa], True
    )
    return ring_atom_df


def get_ring_centroids(ring_atom_df: pd.DataFrame) -> pd.DataFrame:
    """
    Return aromatic ring centroids.

    A helper function for add_aromatic_interactions.

    Computes the ring centroids for each a particular amino acid's ring
    atoms.

    Ring centroids are computed by taking the mean of the x, y, and z
    coordinates.

    Parameters:
    ===========
    - ring_atom_df: a dataframe computed using get_ring_atoms.
    - aa: the amino acid under study
    Returns:
    ========
    - centroid_df: a dataframe containing just the centroid coordinates of
                    the ring atoms of each residue.
    """
    return (
        ring_atom_df.groupby("node_id")
        .mean(numeric_only=True)[["x_coord", "y_coord", "z_coord"]]
        .reset_index()
    )


def compute_ring_normal(ring_df: pd.DataFrame) -> np.ndarray:
    """Compute the normal vector of a ring.

    :param ring_df: Dataframe of atoms in the ring.
    :type ring_df: pd.DataFrame
    :return: Normal vector of the ring.
    :rtype: np.ndarray
    """
    res_name = ring_df["residue_name"].iloc[0]
    atoms = RING_NORMAL_ATOMS[res_name]
    coords = ring_df.loc[ring_df["atom_name"].isin(atoms)][
        ["x_coord", "y_coord", "z_coord"]
    ].values
    pos1, pos2, pos3 = coords[0], coords[1], coords[2]
    return np.cross(pos2 - pos1, pos3 - pos1)


def get_ring_normals(ring_df: pd.DataFrame) -> pd.DataFrame:
    """Compute the normal vector of each ring.

    :param ring_df: Dataframe of atoms in the rings.
    :type ring_df: pd.DataFrame
    :return: Normal vector of the rings.
    :rtype: pd.DataFrame
    """
    return ring_df.groupby("node_id").apply(compute_ring_normal).reset_index()


def compute_angle(
    v1: np.ndarray, v2: np.ndarray, return_degrees: bool = True
) -> float:
    """Computes angle between two vectors.

    :param v1: First vector
    :type v1: np.ndarray
    :param v2: Second vector
    :type v2: np.ndarray
    :param return_degrees: Whether to return angle in degrees or radians
    :type return_degrees: bool
    """
    angle = np.arccos(
        np.dot(v1, v2) / (np.linalg.norm(v1) * np.linalg.norm(v2))
    )
    return 180 * angle / np.pi if return_degrees else angle


def get_edges_by_bond_type(
    G: nx.Graph, bond_type: str
) -> List[Tuple[str, str]]:
    """
    Return edges of a particular bond type.

    Parameters:
    ===========
    - bond_type: (str) one of the elements in the variable BOND_TYPES

    Returns:
    ========
    - resis: (list) a list of tuples, where each tuple is an edge.
    """
    return [
        (n1, n2) for n1, n2, d in G.edges(data=True) if bond_type in d["kind"]
    ]


def node_coords(G: nx.Graph, n: str) -> Tuple[float, float, float]:
    """
    Return the ``x, y, z`` coordinates of a node.
    This is a helper function. Simplifies the code.

    :param G: nx.Graph protein structure graph to extract coordinates from
    :type G: nx.Graph
    :param n: str node ID in graph to extract coordinates from
    :type n: str
    :return: Tuple of coordinates ``(x, y, z)``
    :rtype: Tuple[float, float, float]
    """
    x, y, z = tuple(G.nodes[n]["coords"])
    return x, y, z


def add_interacting_resis(
    G: nx.Graph,
    interacting_atoms: np.ndarray,
    dataframe: pd.DataFrame,
    kind: List[str],
):
    """
    Add interacting residues to graph.

    Returns a list of 2-tuples indicating the interacting residues based
    on the interacting atoms. This is most typically called after the
    get_interacting_atoms function above.

    Also filters out the list such that the residues have to be at least
    two apart.

    ### Parameters

    - interacting_atoms:    (numpy array) result from ``get_interacting_atoms``.
    - dataframe:            (pandas dataframe) a pandas dataframe that
                            houses the euclidean locations of each atom.
    - kind:                 (list) the kind of interaction. Contains one
                            of :
                            - hydrophobic
                            - disulfide
                            - hbond
                            - ionic
                            - aromatic
                            - aromatic_sulphur
                            - cation_pi
                            - delaunay

    Returns:
    ========
    - filtered_interacting_resis: (set of tuples) the residues that are in
        an interaction, with the interaction kind specified
    """
    # This assertion/check is present for defensive programming!
    for k in kind:
        assert k in BOND_TYPES

    resi1 = dataframe.loc[interacting_atoms[0]]["node_id"].values
    resi2 = dataframe.loc[interacting_atoms[1]]["node_id"].values

    interacting_resis = set(list(zip(resi1, resi2)))
    log.info(f"Found {len(interacting_resis)} {k} interactions.")
    for i1, i2 in interacting_resis:
        if i1 != i2:
            if G.has_edge(i1, i2):
                for k in kind:
                    G.edges[i1, i2]["kind"].add(k)
            else:
                G.add_edge(i1, i2, kind=set(kind))
